"""
Represents a single :class:`Document` page, as it would appear in the Textract API output.
The :class:`Page` object also contains the metadata such as the physical dimensions of the page (width, height, in pixels), child_ids etc.
"""

import os
import string
import logging
import xlsxwriter
from typing import List, Tuple
from copy import deepcopy
from collections import defaultdict
from textractor.entities.expense_document import ExpenseDocument

from textractor.entities.word import Word
from textractor.entities.line import Line
from textractor.entities.table import Table
from textractor.entities.signature import Signature
from textractor.entities.layout import Layout
from textractor.entities.page_layout import PageLayout
from textractor.exceptions import InputError
from textractor.entities.key_value import KeyValue
from textractor.entities.query import Query
from textractor.entities.identity_document import IdentityDocument
from textractor.entities.bbox import SpatialObject
from textractor.data.constants import SelectionStatus, Direction, DirectionalFinderType
from textractor.data.constants import TextTypes, SimilarityMetric
from textractor.data.constants import (
    LAYOUT_TEXT,
    LAYOUT_TITLE,
    LAYOUT_HEADER,
    LAYOUT_FOOTER,
    LAYOUT_SECTION_HEADER,
    LAYOUT_PAGE_NUMBER,
    LAYOUT_LIST,
    LAYOUT_FIGURE,
    LAYOUT_TABLE,
    LAYOUT_KEY_VALUE,
)
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.entities.selection_element import SelectionElement
from textractor.utils.geometry_util import sort_by_position
from textractor.utils.search_utils import SearchUtils, jaccard_similarity
from textractor.visualizers.entitylist import EntityList
from textractor.entities.linearizable import Linearizable


class Page(SpatialObject, Linearizable):
    """
    Creates a new document, ideally representing a single item in the dataset.

    :param id: Unique id of the Page
    :type id: str
    :param width: Width of page, in pixels
    :type width: float
    :param height: Height of page, in pixels
    :type height: float
    :param page_num: Page number in the document linked to this Page object
    :type page_num: int
    :param child_ids: IDs of child entities in the Page as determined by Textract
    :type child_ids: List
    """

    def __init__(
        self, id: str, width: int, height: int, page_num: int = -1, child_ids=None
    ):
        super().__init__(width=width, height=height)
        self.id = id
        self._words: EntityList[Word] = EntityList([])
        self._lines: EntityList[Line] = EntityList([])
        self._key_values: EntityList[KeyValue] = EntityList([])
        self._checkboxes: EntityList[KeyValue] = EntityList([])
        self._tables: EntityList[Table] = EntityList([])
        self._queries: EntityList[Query] = EntityList([])
        self._expense_documents: EntityList[ExpenseDocument] = EntityList([])
        self._leaf_layouts: EntityList[Layout] = EntityList([])
        self._container_layouts: EntityList[Layout] = EntityList([])
        self.kv_cache = defaultdict(list)
        self.metadata = {}
        self.page_num = page_num
        self.child_ids: List[str] = child_ids
        self.image = None

    # functions to add entities to the Page object
    @property
    def words(self) -> EntityList[Word]:
        """
        Returns all the :class:`Word` objects present in the Page.

        :return: List of Word objects, each representing a word within the Page.
        :rtype: EntityList[Word]
        """
        return self._words

    @words.setter
    def words(self, words: List[Word]):
        """
        Add Word objects to the Page.

        :param words: List of Word objects, each representing a word within the page. No specific ordering is assumed.
        :type words: List[Word]
        """
        self._words = words
        self._words = EntityList(sort_by_position(list(set(self._words))))

    @property
    def lines(self) -> EntityList[Line]:
        """
        Returns all the :class:`Line` objects present in the Page.

        :return: List of Line objects, each representing a line within the Page.
        :rtype: EntityList[Line]
        """
        return self._lines

    @lines.setter
    def lines(self, lines: List[Line]):
        """
        Add Line objects to the Document.

        :param lines: List of Line objects, each representing a line within the Page.
        :type lines: List[Line]
        """
        self._lines = EntityList(sort_by_position(lines))

    @property
    def text(self) -> str:
        """
        Returns the page text

        :return: Linearized page text
        :rtype: str
        """
        return self.get_text()

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> Tuple[str, List[Word]]:
        """
        Returns the page text and words sorted in reading order

        :param config: Text linearization configuration object, defaults to TextLinearizationConfig()
        :type config: TextLinearizationConfig, optional
        :return: Tuple of page text and words
        :rtype: Tuple[str, List[Word]]
        """
        unsorted_layouts = [l for l in self.layouts if l.reading_order < 0]
        sorted_layouts = [l for l in self.layouts if l.reading_order >= 0]
        if unsorted_layouts:
            for unsorted_layout in sorted(
                unsorted_layouts, key=lambda x: (x.bbox.y, x.bbox.x)
            ):
                closest_layout = None
                closest_reading_order_distance = None
                for layout in sorted(sorted_layouts, key=lambda x: x.reading_order):
                    dist = layout.bbox.get_distance(unsorted_layout)
                    if (
                        closest_reading_order_distance is None
                        or dist < closest_reading_order_distance
                    ):
                        closest_layout = layout
                if closest_layout:
                    sorted_layouts.insert(
                        sorted_layouts.index(closest_layout) + 1, unsorted_layout
                    )
                else:
                    sorted_layouts.append(unsorted_layout)

        page_texts_and_words = [l.get_text_and_words(config) for l in sorted_layouts]

        if not page_texts_and_words:
            return "", []

        text, words = zip(*page_texts_and_words)
        combined_words = []
        for w in words:
            combined_words += w
        return config.layout_element_separator.join(text), combined_words

    @property
    def page_layout(self) -> PageLayout:
        return PageLayout(
            titles=EntityList(
                [l for l in self.layouts if l.layout_type == LAYOUT_TITLE]
            ),
            headers=EntityList(
                [l for l in self.layouts if l.layout_type == LAYOUT_HEADER]
            ),
            footers=EntityList(
                [l for l in self.layouts if l.layout_type == LAYOUT_FOOTER]
            ),
            section_headers=EntityList(
                [l for l in self.layouts if l.layout_type == LAYOUT_SECTION_HEADER]
            ),
            page_numbers=EntityList(
                [l for l in self.layouts if l.layout_type == LAYOUT_PAGE_NUMBER]
            ),
            lists=EntityList([l for l in self.layouts if l.layout_type == LAYOUT_LIST]),
            figures=EntityList(
                [l for l in self.layouts if l.layout_type == LAYOUT_FIGURE]
            ),
            tables=EntityList(
                [l for l in self.layouts if l.layout_type == LAYOUT_TABLE]
            ),
            key_values=EntityList(
                [l for l in self.layouts if l.layout_type == LAYOUT_KEY_VALUE]
            ),
        )

    @property
    def key_values(self) -> EntityList[KeyValue]:
        """
        Returns all the :class:`KeyValue` objects present in the Page.

        :return: List of KeyValue objects, each representing a key-value pair within the Page.
        :rtype: EntityList[KeyValue]
        """
        return self._key_values

    @key_values.setter
    def key_values(self, kv: List[KeyValue]):
        """
        Add KeyValue objects to the Page.

        :param kv: List of KeyValue objects, each representing a KV area within the document page.
        :type kv: List[KeyValue]
        """
        self._key_values = EntityList(sort_by_position(kv))

    @property
    def checkboxes(self) -> EntityList[KeyValue]:
        """
        Returns all the :class:`KeyValue` objects with :class:`SelectionElement` present in the Page.

        :return: List of KeyValue objects, each representing a checkbox within the Page.
        :rtype: EntityList[KeyValue]
        """
        return self._checkboxes

    @checkboxes.setter
    def checkboxes(self, checkbox: List[KeyValue]):
        """
        Add KeyValue objects containing SelectionElement children to the Page.

        :param checkbox: List of KeyValue objects, each representing a checkbox area within the document page.
        :type checkbox: List[KeyValue]
        """
        self._checkboxes = EntityList(sort_by_position(checkbox))

    @property
    def tables(self) -> EntityList[Table]:
        """
        Returns all the :class:`Table` objects present in the Page.

        :return: List of Table objects, each representing a table within the Page.
        :rtype: EntityList
        """
        return self._tables

    @tables.setter
    def tables(self, tables: List[Table]):
        """
        Add Table objects to the Page.

        :param tables: List of Table objects, each representing a Table area within the document page.
        :type tables: list
        """
        self._tables = EntityList(tables)

    @property
    def queries(self) -> EntityList[Query]:
        """
        Returns all the :class:`Query` objects present in the Page.

        :return: List of Query objects.
        :rtype: EntityList
        """
        return self._queries

    @queries.setter
    def queries(self, queries: List[Query]):
        """
        Add Signature objects to the Page.

        :param signatures: List of Signature objects.
        :type signatures: list
        """
        self._queries = EntityList(queries)

    @property
    def signatures(self) -> EntityList[Signature]:
        """
        Returns all the :class:`Signature` objects present in the Page.

        :return: List of Signature objects.
        :rtype: EntityList
        """
        return self._signatures

    @signatures.setter
    def signatures(self, signatures: List[Signature]):
        """
        Add Signature objects to the Page.

        :param signatures: List of Signature objects.
        :type signatures: list
        """
        self._signatures = EntityList(signatures)

    @property
    def layouts(self) -> EntityList[Layout]:
        """
        Returns all the :class:`Layout` objects present in the Page.

        :return: List of Layout objects.
        :rtype: EntityList
        """
        return EntityList(
            sorted(
                self._leaf_layouts + self._container_layouts,
                key=lambda c: c.reading_order,
            )
        )

    @property
    def leaf_layouts(self) -> EntityList[Layout]:
        """
        Returns all the leaf :class:`Layout` objects present in the Page.

        :return: List of Layout objects.
        :rtype: EntityList
        """
        return self._leaf_layouts

    @leaf_layouts.setter
    def leaf_layouts(self, leaf_layouts: List[Layout]):
        """
        Add leaf layout objects to the Page.

        :param layouts: List of Layout objects.
        :type layouts: list
        """
        self._leaf_layouts = EntityList(leaf_layouts)

    @property
    def container_layouts(self) -> EntityList[Layout]:
        """
        Returns all the container :class:`Layout` objects present in the Page.

        :return: List of Layout objects.
        :rtype: EntityList
        """
        return self._container_layouts

    @container_layouts.setter
    def container_layouts(self, container_layouts: List[Layout]):
        """
        Add Layout objects to the Page.

        :param layouts: List of Layout objects.
        :type layouts: list
        """
        self._container_layouts = EntityList(container_layouts)

    @property
    def expense_documents(self) -> EntityList[ExpenseDocument]:
        """
        Returns all the :class:`ExpenseDocument` objects present in the Page.

        :return: List of ExpenseDocument objects.
        :rtype: EntityList
        """
        return self._expense_documents

    @expense_documents.setter
    def expense_documents(self, expense_documents: List[ExpenseDocument]):
        """
        Add ExpenseDocument objects to the Page.

        :param tables: List of ExpenseDocument objects.
        :type expense_documents: list
        """
        self._expense_documents = EntityList(expense_documents)

    def __repr__(self):
        return os.linesep.join(
            [
                f"This Page ({self.page_num}) holds the following data:",
                f"Words - {len(self.words)}",
                f"Lines - {len(self.lines)}",
                f"Key-values - {len(self.key_values)}",
                f"Checkboxes - {len(self.checkboxes)}",
                f"Tables - {len(self.tables)}",
                f"Queries - {len(self.queries)}",
                f"Signatures - {len(self.signatures)}",
                f"Expense documents - {len(self.expense_documents)}",
                f"Layout elements - {len(self.layouts)}",
            ]
        )

    def __getitem__(self, key):
        output = self.get(key)
        if output:
            return output
        raise KeyError(f"{key} was not found in Document")

    def keys(self, include_checkboxes: bool = True) -> List[str]:
        """
        Prints all keys for key-value pairs and checkboxes if the page contains them.

        :param include_checkboxes: True/False. Set False if checkboxes need to be excluded.
        :type include_checkboxes: bool

        :return: List of strings containing key names in the Page
        :rtype: List[str]
        """
        keys = []
        keys = [keyvalue.key for keyvalue in self.key_values]
        if include_checkboxes:
            keys += [keyvalue.key for keyvalue in self.checkboxes]
        return keys

    def filter_checkboxes(
        self, selected: bool = True, not_selected: bool = True
    ) -> EntityList[KeyValue]:
        """
        Return a list of :class:`KeyValue` objects containing checkboxes if the page contains them.

        :param selected: True/False Return SELECTED checkboxes
        :type selected: bool
        :param not_selected: True/False Return NOT_SELECTED checkboxes
        :type not_selected: bool

        :return: Returns checkboxes that match the conditions set by the flags.
        :rtype: EntityList[KeyValue]
        """
        if not self.checkboxes:
            logging.warning(f"This document does not contain checkboxes")
            return []
        else:
            if selected and not_selected:
                checkboxes = self.checkboxes
                return EntityList(checkboxes)

            checkboxes = []
            if selected:
                checkboxes = [
                    kv
                    for kv in self.checkboxes
                    if kv.selection_status == SelectionStatus.SELECTED
                ]
            if not_selected:
                checkboxes = [
                    kv
                    for kv in self.checkboxes
                    if kv.selection_status == SelectionStatus.NOT_SELECTED
                ]

            return EntityList(checkboxes)

    def get_words_by_type(
        self, text_type: TextTypes = TextTypes.PRINTED
    ) -> EntityList[Word]:
        """
        Returns list of :class:`Word` entities that match the input text type.

        :param text_type: TextTypes.PRINTED or TextTypes.HANDWRITING
        :type text_type: TextTypes

        :return: Returns list of Word entities that match the input text type.
        :rtype: EntityList[Word]
        """
        if not isinstance(text_type, TextTypes):
            raise InputError(
                "text_type parameter should be of TextTypes type. Find input choices from textractor.data.constants"
            )

        if not self.words:
            logging.warn("Document contains no word entities.")
            return []

        filtered_words = [word for word in self.words if word.text_type == text_type]
        return EntityList(filtered_words)

    def _search_words_with_similarity(
        self,
        keyword: str,
        top_k: int = 1,
        similarity_metric: SimilarityMetric = SimilarityMetric.LEVENSHTEIN,
        similarity_threshold: float = 0.6,
    ) -> List[Tuple[Word, float]]:
        """
        Returns a list of top_k words with their similarity to the keyword.

        :param keyword: Keyword that is used to query the document.
        :type keyword: str, required
        :param top_k: Number of closest word objects to be returned. default=1
        :type top_k: int, optional
        :param similarity_metric: SimilarityMetric.COSINE, SimilarityMetric.EUCLIDEAN or SimilarityMetric.LEVENSHTEIN. SimilarityMetric.COSINE is chosen as default.
        :type similarity_metric: SimilarityMetric
        :param similarity_threshold: Measure of how similar document key is to queried key. default=0.6
        :type similarity_threshold: float

        :return: Returns a list of tuples containing similarity and Word.
        :rtype: List[Tuple(float, Word)]]
        """
        if not isinstance(similarity_metric, SimilarityMetric):
            raise InputError(
                "similarity_metric parameter should be of SimilarityMetric type. Find input choices from textractor.data.constants"
            )
        top_n_words = []
        similarity_threshold = (
            similarity_threshold
            if similarity_metric == SimilarityMetric.COSINE
            else -(similarity_threshold)
        )
        lowest_similarity = similarity_threshold

        for word in self.words:
            similarity = SearchUtils.get_word_similarity(
                keyword, word.text, similarity_metric
            )
            similarity = (
                similarity
                if similarity_metric == SimilarityMetric.COSINE
                else -(similarity)
            )

            if len(top_n_words) < top_k and similarity > similarity_threshold:
                top_n_words.append((similarity, word))
            elif similarity > lowest_similarity:
                top_n_words[-1] = (similarity, word)
            else:
                continue
            top_n_words = sorted(top_n_words, key=lambda x: x[0], reverse=True)
            lowest_similarity = top_n_words[-1][0]

        return top_n_words

    def search_words(
        self,
        keyword: str,
        top_k: int = 1,
        similarity_metric: SimilarityMetric = SimilarityMetric.LEVENSHTEIN,
        similarity_threshold: float = 0.6,
    ) -> EntityList[Word]:
        """
        Return a list of top_k words that match the keyword.

        :param keyword: Keyword that is used to query the document.
        :type keyword: str, required
        :param top_k: Number of closest word objects to be returned. default=1
        :type top_k: int, optional
        :param similarity_metric: SimilarityMetric.COSINE, SimilarityMetric.EUCLIDEAN or SimilarityMetric.LEVENSHTEIN. SimilarityMetric.COSINE is chosen as default.
        :type similarity_metric: SimilarityMetric
        :param similarity_threshold: Measure of how similar document key is to queried key. default=0.6
        :type similarity_threshold: float

        :return: Returns a list of words that match the queried key sorted from highest to lowest similarity.
        :rtype: EntityList[Word]
        """

        top_n_words = EntityList(
            [
                ent[1]
                for ent in self._search_words_with_similarity(
                    keyword=keyword,
                    top_k=top_k,
                    similarity_metric=similarity_metric,
                    similarity_threshold=similarity_threshold,
                )
            ]
        )

        return top_n_words

    def _search_lines_with_similarity(
        self,
        keyword: str,
        top_k: int = 1,
        similarity_metric: SimilarityMetric = SimilarityMetric.LEVENSHTEIN,
        similarity_threshold: int = 0.6,
    ) -> List[Tuple[Line, float]]:
        """
        Return a list of top_k lines that contain the queried keyword.

        :param keyword: Keyword that is used to query the page.
        :type keyword: str
        :param top_k: Number of closest line objects to be returned
        :type top_k: int
        :param similarity_metric: SimilarityMetric.COSINE, SimilarityMetric.EUCLIDEAN or SimilarityMetric.LEVENSHTEIN. SimilarityMetric.COSINE is chosen as default.
        :type similarity_metric: SimilarityMetric
        :param similarity_threshold: Measure of how similar page key is to queried key. default=0.6
        :type similarity_threshold: float

        :return: Returns a list of tuples of lines and their similarity to the keyword that contain the queried key sorted
                 from highest to lowest similarity.
        :rtype: List[Tuple[Line, float]]
        """
        if not isinstance(similarity_metric, SimilarityMetric):
            raise InputError(
                "similarity_metric parameter should be of SimilarityMetric type. Find input choices from textractor.data.constants"
            )

        top_n_lines = []
        similarity_threshold = (
            similarity_threshold
            if similarity_metric == SimilarityMetric.COSINE
            else -(similarity_threshold)
        )
        lowest_similarity = similarity_threshold

        for line in self.lines:
            similarity = [
                SearchUtils.get_word_similarity(keyword, word, similarity_metric)
                for word in line.__repr__().split(" ")
            ]
            similarity.append(
                SearchUtils.get_word_similarity(
                    keyword, line.__repr__(), similarity_metric
                )
            )
            similarity = (
                max(similarity)
                if similarity_metric == SimilarityMetric.COSINE
                else -min(similarity)
            )

            if len(top_n_lines) < top_k and similarity > similarity_threshold:
                top_n_lines.append((similarity, line))
            elif similarity > lowest_similarity:
                top_n_lines[-1] = (similarity, line)
            else:
                continue
            top_n_lines = sorted(top_n_lines, key=lambda x: x[0], reverse=True)
            lowest_similarity = top_n_lines[-1][0]

        return top_n_lines

    def search_lines(
        self,
        keyword: str,
        top_k: int = 1,
        similarity_metric: SimilarityMetric = SimilarityMetric.LEVENSHTEIN,
        similarity_threshold: int = 0.6,
    ) -> EntityList[Line]:
        """
        Return a list of top_k lines that contain the queried keyword.

        :param keyword: Keyword that is used to query the page.
        :type keyword: str
        :param top_k: Number of closest line objects to be returned
        :type top_k: int
        :param similarity_metric: SimilarityMetric.COSINE, SimilarityMetric.EUCLIDEAN or SimilarityMetric.LEVENSHTEIN. SimilarityMetric.COSINE is chosen as default.
        :type similarity_metric: SimilarityMetric
        :param similarity_threshold: Measure of how similar page key is to queried key. default=0.6
        :type similarity_threshold: float

        :return: Returns a list of lines that contain the queried key sorted
                 from highest to lowest similarity.
        :rtype: EntityList[Line]
        """

        top_n_lines = EntityList(
            [
                ent[1]
                for ent in self._search_lines_with_similarity(
                    keyword=keyword,
                    top_k=top_k,
                    similarity_metric=similarity_metric,
                    similarity_threshold=similarity_threshold,
                )
            ]
        )

        return top_n_lines

    # KeyValue entity related functions
    def get(
        self,
        key: str,
        top_k_matches: int = 1,
        similarity_metric: SimilarityMetric = SimilarityMetric.LEVENSHTEIN,
        similarity_threshold: float = 0.6,
    ) -> EntityList[KeyValue]:
        """
        Return upto top_k_matches of key-value pairs for the key that is queried from the page.

        :param key: Query key to match
        :type key: str
        :param top_k_matches: Maximum number of matches to return
        :type top_k_matches: int
        :param similarity_metric: SimilarityMetric.COSINE, SimilarityMetric.EUCLIDEAN or SimilarityMetric.LEVENSHTEIN. SimilarityMetric.COSINE is chosen as default.
        :type similarity_metric: SimilarityMetric
        :param similarity_threshold: Measure of how similar page key is to queried key. default=0.6
        :type similarity_threshold: float

        :return: Returns a list of key-value pairs that match the queried key sorted from highest
                 to lowest similarity.
        :rtype: EntityList[KeyValue]
        """
        if not isinstance(similarity_metric, SimilarityMetric):
            raise InputError(
                "similarity_metric parameter should be of SimilarityMetric type. Find input choices from textractor.data.constants"
            )

        top_n = []
        similarity_threshold = (
            similarity_threshold
            if similarity_metric == SimilarityMetric.COSINE
            else -(similarity_threshold)
        )
        lowest_similarity = similarity_threshold

        for kv in self.key_values + self.checkboxes:
            try:
                edited_document_key = "".join(
                    [
                        char
                        for char in kv.key.__repr__()
                        if char not in string.punctuation
                    ]
                )
            except:
                pass
            key = "".join([char for char in key if char not in string.punctuation])

            similarity = [
                SearchUtils.get_word_similarity(key, word, similarity_metric)
                for word in edited_document_key.split(" ")
            ]
            similarity.append(
                SearchUtils.get_word_similarity(
                    key, edited_document_key, similarity_metric
                )
            )

            similarity = (
                max(similarity)
                if similarity_metric == SimilarityMetric.COSINE
                else -min(similarity)
            )

            if similarity > similarity_threshold:
                if len(top_n) < top_k_matches:
                    top_n.append((kv, similarity))
                elif similarity > lowest_similarity:
                    top_n[-1] = (kv, similarity)
                top_n = sorted(top_n, key=lambda x: x[1], reverse=True)
                lowest_similarity = top_n[-1][1]

        if not top_n:
            logging.warning(
                f"Query key does not match any existing keys in the document.{os.linesep}{self.keys()}"
            )

        logging.info(f"Query key matched {len(top_n)} key-values in the document.")

        return EntityList([value[0] for value in top_n])

    # Export document entities into supported formats
    def export_kv_to_csv(
        self,
        include_kv: bool = True,
        include_checkboxes: bool = True,
        filepath: str = "Key-Values.csv",
    ):
        """
        Export key-value entities and checkboxes in csv format.

        :param include_kv: True if KVs are to be exported. Else False.
        :type include_kv: bool
        :param include_checkboxes: True if checkboxes are to be exported. Else False.
        :type include_checkboxes: bool
        :param filepath: Path to where file is to be stored.
        :type filepath: str
        """
        keys = []
        values = []
        if include_kv and not self.key_values:
            logging.warning("Document does not contain key-values.")
        elif include_kv:
            for kv in self.key_values:
                keys.append(kv.key.__repr__())
                values.append(kv.value.__repr__())

        if include_checkboxes and not self.checkboxes:
            logging.warning("Document does not contain checkbox elements.")
        elif include_checkboxes:
            for kv in self.checkboxes:
                keys.append(kv.key.__repr__())
                values.append(kv.value.children[0].status.name)

        with open(filepath, "w") as f:
            f.write(f"Key,Value{os.linesep}")
            for k, v in zip(keys, values):
                f.write(f"{k},{v}{os.linesep}")

        logging.info(
            f"csv file stored at location {os.path.join(os.getcwd(), filepath)}"
        )

    def export_kv_to_txt(
        self,
        include_kv: bool = True,
        include_checkboxes: bool = True,
        filepath: str = "Key-Values.txt",
    ):
        """
        Export key-value entities and checkboxes in txt format.

        :param include_kv: True if KVs are to be exported. Else False.
        :type include_kv: bool
        :param include_checkboxes: True if checkboxes are to be exported. Else False.
        :type include_checkboxes: bool
        :param filepath: Path to where file is to be stored.
        :type filepath: str
        """
        export_str = []
        index = 1
        if include_kv and not self.key_values:
            logging.warning("Document does not contain key-values.")
        elif include_kv:
            for kv in self.key_values:
                export_str.append(
                    f"{index}. {kv.key.__repr__()} : {kv.value.__repr__()}{os.linesep}"
                )
                index += 1

        if include_checkboxes and not self.checkboxes:
            logging.warning("Document does not contain checkbox elements.")
        elif include_checkboxes:
            for kv in self.checkboxes:
                export_str.append(
                    f"{index}. {kv.key.__repr__()} : {kv.value.__repr__()}{os.linesep}"
                )
                index += 1

        with open(filepath, "w") as text_file:
            text_file.write("".join(export_str))
        logging.info(
            f"txt file stored at location {os.path.join(os.getcwd(),filepath)}"
        )

    def independent_words(self) -> EntityList[Word]:
        """
        :return: Return all words in the document, outside of tables, checkboxes, key-values.
        :rtype: EntityList[Word]
        """
        if not self.words:
            logging.warning("Words have not been assigned to this Document object.")
            return []

        else:
            table_words = sum([table.words for table in self.tables], [])
            kv_words = sum([kv.words for kv in self.key_values], [])
            checkbox_words = sum([kv.words for kv in self.checkboxes], [])
            dependent_words = table_words + checkbox_words + kv_words
            dependent_word_ids = set([word.id for word in dependent_words])
            independent_words = [
                word for word in self.words if word.id not in dependent_word_ids
            ]
            return EntityList(independent_words)

    def export_tables_to_excel(self, filepath):
        """
        Creates an excel file and writes each table on a separate worksheet within the workbook.
        This is stored on the filepath passed by the user.

        :param filepath: Path to store the exported Excel file.
        :type filepath: str, required
        """
        if not filepath:
            logging.error("Filepath required to store excel file.")
        workbook = xlsxwriter.Workbook(filepath)
        for table in self.tables:
            workbook = table.to_excel(
                filepath=None, workbook=workbook, save_workbook=False
            )
        workbook.close()

    def _update_entity_page_num(self):
        """Updates page number if Textractor API call was given a list of images."""
        entities = (
            self.words + self.lines + self.key_values + self.checkboxes + self.tables
        )
        for entity in entities:
            entity.page = self.page_num

    def return_duplicates(self):
        """
        Returns a list containing :class:`EntityList` objects.
        Each :class:`EntityList` instance contains the key-values and the last item is the table which contains duplicate information.
        This function is intended to let the Textract user know of duplicate objects extracted by the various Textract models.

        :return: List of EntityList objects each containing the intersection of KeyValue and Table entities on the page.
        :rtype: List[EntityList]
        """
        tables = self.tables
        key_values = self.key_values

        document_duplicates = []

        for table in tables:
            table_duplicates = EntityList([])
            table_x1, table_x2, table_y1, table_y2 = (
                table.bbox.x,
                table.bbox.x + table.bbox.width,
                table.bbox.y,
                table.bbox.y + table.bbox.height,
            )
            for kv in key_values:
                if (
                    kv.bbox.x >= table_x1
                    and kv.bbox.x <= table_x2
                    and kv.bbox.y >= table_y1
                    and kv.bbox.y <= table_y2
                ):
                    table_duplicates.append(kv)

            if table_duplicates:
                table_duplicates.append(table)

            document_duplicates.append(table_duplicates)

        return document_duplicates

    def directional_finder(
        self,
        word_1: str = "",
        word_2: str = "",
        prefix: str = "",
        direction=Direction.BELOW,
        entities=[],
    ):
        """
        The function returns entity types present in entities by prepending the prefix provided by te user. This helps in cases of repeating
        key-values and checkboxes. The user can manipulate original data or produce a copy. The main advantage of this function is to be able to define direction.

        :param word_1: The reference word from where x1, y1 coordinates are derived
        :type word_1: str, required
        :param word_2: The second word preferably in the direction indicated by the parameter direction. When it isn't given the end of page coordinates are used in the given direction.
        :type word_2: str, optional
        :param prefix: User provided prefix to prepend to the key . Without prefix, the method acts as a search by geometry function
        :type prefix: str, optional
        :param entities: List of DirectionalFinderType inputs.
        :type entities: List[DirectionalFinderType]

        :return: Returns the EntityList of modified key-value and/or checkboxes
        :rtype: EntityList
        """

        if not word_1:
            return EntityList([])

        x1, x2, y1, y2 = self._get_coords(word_1, word_2, direction)

        if x1 == -1:
            return EntityList([])

        entity_dict = {
            DirectionalFinderType.KEY_VALUE_SET: self.key_values,
            DirectionalFinderType.SELECTION_ELEMENT: self.checkboxes,
        }

        entitylist = []
        for entity_type in entities:
            entitylist.extend(list(entity_dict[entity_type]))

        new_key_values = self._get_kv_with_direction(
            direction, entitylist, (x1, x2, y1, y2)
        )

        final_kv = []
        for kv in new_key_values:
            if kv.key:
                key_words = [deepcopy(word) for word in kv.key]
                key_words[0].text = prefix + key_words[0].text
                new_kv = deepcopy(kv)
                new_kv.key = key_words
                final_kv.append(new_kv)
            else:
                final_kv.append(kv)

        return EntityList(final_kv)

    def _get_kv_with_direction(self, direction, entitylist, coords):
        """Return key-values and checkboxes in entitiylist present in the direction given with respect to the coordinates."""
        if direction == Direction.ABOVE:
            new_key_values = [
                kv
                for kv in entitylist
                if kv.bbox.y <= coords[2] and kv.bbox.y >= coords[-1]
            ]

        elif direction == Direction.BELOW:
            new_key_values = [
                kv
                for kv in entitylist
                if kv.bbox.y >= coords[2] and kv.bbox.y <= coords[-1]
            ]

        elif direction == Direction.RIGHT:
            new_key_values = [
                kv
                for kv in entitylist
                if kv.bbox.x >= coords[0] and kv.bbox.x <= coords[1]
            ]
            new_key_values = [
                kv
                for kv in new_key_values
                if kv.bbox.y >= coords[2] - kv.bbox.height
                and kv.bbox.y <= coords[-1] + 3 * kv.bbox.height
            ]

        elif direction == Direction.LEFT:
            new_key_values = [
                kv
                for kv in entitylist
                if kv.bbox.x <= coords[0] and kv.bbox.x >= coords[1]
            ]
            new_key_values = [
                kv
                for kv in new_key_values
                if kv.bbox.y >= coords[2] - kv.bbox.height
                and kv.bbox.y <= coords[-1] + 3 * kv.bbox.height
            ]

        return new_key_values

    def _get_coords(self, word_1, word_2, direction):
        """
        Returns coordinates for the area within which to search for key-values with the directional_finder by retrieving coordinates of word_1 \
        and word_2 if it exists else end of page.
        """
        word_1_objects = self.search_lines(
            keyword=word_1,
            top_k=5,
            similarity_metric=SimilarityMetric.COSINE,
            similarity_threshold=0.5,
        )

        if not word_1_objects:
            logging.warning(f"{word_1} not found in page")
            return -1, -1, -1, -1
        else:
            word_1_obj = word_1_objects[0]
            x1, y1 = word_1_obj.bbox.x, word_1_obj.bbox.y

        if word_2:
            word_2_objects = self.search_lines(
                keyword=word_2,
                top_k=5,
                similarity_metric=SimilarityMetric.COSINE,
                similarity_threshold=0.5,
            )

            if not word_2_objects:
                logging.warning(f"{word_2} not found in page")
                return -1, -1, -1, -1
            else:
                word_2_obj = word_2_objects[0]
                x2, y2 = word_2_obj.bbox.x, word_2_obj.bbox.y
        else:
            x2, y2 = 1, 1

        if direction == Direction.ABOVE:
            x1, x2, y1, y2 = (x1, x2, y1, y2) if y2 < y1 else (x1, 0, y1, 0)

        elif direction == Direction.BELOW:
            x1, x2, y1, y2 = (x1, x2, y1, y2) if y2 > y1 else (x1, 1, y1, 1)

        elif direction == Direction.RIGHT:
            x1, x2, y1, y2 = (x1, x2, y1, y2) if x2 > x1 else (x1, 1, y1, y1)

        elif direction == Direction.LEFT:
            x1, x2, y1, y2 = (x1, x2, y1, y2) if x2 < x1 else (x1, 0, y1, y1)

        else:
            return -1, -1, -1, -1

        return x1, x2, y1, y2

    def visualize(self, *args, **kwargs):
        """
        Returns the object's children in a visualization EntityList object

        :return: Returns an EntityList object
        :rtype: EntityList
        """
        return EntityList(self).visualize(*args, **kwargs)



"""
The :class:`KeyValue` entity is a document entity representing the Forms output. The key in :class:`KeyValue` are typically words 
and the :class:`Value` could be :class:`Word` elements or :class:`SelectionElement` in case of checkboxes.

This class contains the associated metadata with the :class:`KeyValue` entity including the entity ID, 
bounding box information, value, existence of checkbox, page number, Page ID and confidence of detection.
"""

from typing import List, Optional, Tuple
from textractor.data.text_linearization_config import TextLinearizationConfig

from textractor.entities.query_result import QueryResult
from textractor.entities.bbox import BoundingBox
from textractor.entities.document_entity import DocumentEntity


class Query(DocumentEntity):
    """
    The Query object merges QUERY and QUERY_RESULT blocks.
    To create a new :class:`Query` object we require the following:

    :param entity_id: Unique identifier of the Query entity.
    :type entity_id: str
    :param bbox: Bounding box of the KeyValue entity.
    :type bbox: BoundingBox
    :param contains_checkbox: True/False to indicate if the value is a checkbox.
    :type contains_checkbox: bool
    :param value: Value object that maps to the KeyValue entity.
    :type value: Value
    :param confidence: confidence with which the entity was detected.
    :type confidence: float
    """

    def __init__(
        self,
        entity_id: str,
        query: str,
        alias: str,
        query_result: Optional[QueryResult],
        result_bbox: Optional[BoundingBox],
    ):
        super().__init__(entity_id, result_bbox)

        self.query = query
        self.alias = alias
        self.result = query_result
        self._page = None
        self._page_id = None

    @property
    def page(self) -> int:
        """
        :return: Returns the page number of the page the :class:`Table` entity is present in.
        :rtype: int
        """
        return self._page

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the :class:`Table` entity.

        :param page_num: Page number where the Table entity exists.
        :type page_num: int
        """
        self._page = page_num

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """
        return self._page_id

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the :class:`Table` entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """
        self._page_id = page_id

    @property
    def has_result(self) -> bool:
        """
        :return: Returns whether there was a result associated with the query
        :rtype: bool
        """
        return self.result is not None

    def __repr__(self) -> str:
        """
        :return: Returns Query object as a formatted string.
        :rtype: str
        """

        if self.result:
            return f"{self.query} {self.result.answer}"
        else:
            return f"{self.query}"

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> Tuple[str, List]:
        """
        Used for linearization, returns the linearized text of the Query and the matching words

        :return: Tuple of text and word list
        :rtype: Tuple[str, List[Word]]
        """
        return f"{self.query} {self.result.answer}", []



import os
from dataclasses import dataclass

from textractor.data.text_linearization_config import TextLinearizationConfig

@dataclass
class MarkdownLinearizationConfig(TextLinearizationConfig):
    """
    This :class:`MarkdownLinearizationConfig` is a convenience configuration for converting a Document or DocumentEntity to Markdown.
    For a description of the parameters see :class:`TextLinearizationConfig`.
    """

    title_prefix: str = "# "

    table_linearization_format: str = "markdown"

    section_header_prefix: str = "## "

    table_remove_column_headers: bool = True


"""The IdentityDocument class is the object representation of an AnalyzeID response. It is similar to a dictionary. Despite its name it does not inherit from Document as the AnalyzeID response does not contains position information."""

import os
import string
import logging
import xlsxwriter
from typing import List, Dict, Union
from copy import deepcopy
from collections import defaultdict
from textractor.data.constants import AnalyzeIDFields
from textractor.entities.bbox import SpatialObject
from textractor.entities.identity_field import IdentityField

from textractor.exceptions import InputError


class IdentityDocument(SpatialObject):
    """
    Represents the description of a single ID document.
    """

    def __init__(self, fields=None):
        """
        Creates a new document, ideally containing entity objects pertaining to each page.

        :param num_pages: Number of pages in the input Document.
        """
        super().__init__(width=0, height=0)
        self._fields = IdentityDocument._fields_to_dict(fields)

    @classmethod
    def _fields_to_dict(cls, fields: Union[List[IdentityField], Dict[str, dict]]):
        if not fields:
            return {}
        elif isinstance(fields, list) and isinstance(fields[0], IdentityField):
            return {id_field.key: id_field for id_field in fields}
        elif isinstance(fields, dict):
            field_dict = {}
            for id_field in fields.values():
                field_dict[id_field["key"]] = IdentityField(
                    id_field["key"],
                    id_field["value"],
                    id_field["confidence"],
                )
            return field_dict
        else:
            raise InputError(
                f"fields needs to be a list of IdentityFields or a list of dictionaries, not {type(fields)}"
            )

    @property
    def fields(self) -> Dict[str, IdentityField]:
        return self._fields

    @fields.setter
    def fields(self, fields):
        self._fields = fields

    def keys(self) -> List[str]:
        keys = [key for key in self._fields.keys()]
        return keys

    def values(self) -> List[str]:
        values = [field.value for field in self._fields.values()]
        return values

    def __getitem__(self, key: Union[str, AnalyzeIDFields]) -> str:
        return self._fields[key if isinstance(key, str) else key.value].value

    def get(self, key: Union[str, AnalyzeIDFields]) -> Union[str, None]:
        result = self._fields.get(key if isinstance(key, str) else key.value)
        if result is None:
            return None
        return result.value

    def __repr__(self):
        return os.linesep.join([f"{str(k)}: {str(v)}" for k, v in self.fields.items()])



"""
Represents a single :class:`TableTitle:class:` object. The `TableCell:class:` object contains information such as:

* The position of the title within the Document
* The words that it contains
* Confidence of entity detection
"""

from typing import List
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.entities.bbox import BoundingBox
from textractor.entities.document_entity import DocumentEntity
from textractor.entities.word import Word
from textractor.utils.text_utils import linearize_children
from textractor.visualizers.entitylist import EntityList


class TableTitle(DocumentEntity):
    """
    Represents a title that is either in-table or floating
    """

    def __init__(
        self,
        entity_id: str,
        bbox: BoundingBox,
    ):
        super().__init__(entity_id, bbox)
        self._words: List[Word] = []
        self._is_floating: bool = False
        self._page = None
        self._page_id = None

    @property
    def words(self):
        """
        Returns all the Word objects present in the :class:`TableTitle`.

        :return words: List of Word objects, each representing a word within the TableTitle.
        :rtype: list
        """
        return EntityList(self._words)

    @words.setter
    def words(self, words: List[Word]):
        """
        Add Word objects to the :class:`TableTitle`.

        :param words: List of Word objects, each representing a word within the TableTitle. No specific ordering is assumed as it is ordered internally.
        :type words: list
        """
        self._words = words

    @property
    def text(self) -> str:
        """Returns the text in the title as one space-separated string

        :return: Text in the title
        :rtype: str
        """
        return " ".join([w.text for w in self.words])

    @property
    def page(self):
        """
        :return: Returns the page number of the page the TableTitle entity is present in.
        :rtype: int
        """

        return self._page

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the TableTitle entity.

        :param page_num: Page number where the TableTitle entity exists.
        :type page_num: int
        """

        self._page = page_num

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """

        return self._page_id

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the TableTitle entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """

        self._page_id = page_id

    @property
    def is_floating(self) -> bool:
        """
        :return: Returns whether the TableTitle entity is floating or not.
        :rtype: bool
        """

        return self._is_floating

    @is_floating.setter
    def is_floating(self, is_floating: bool):
        """
        Sets the is_floating attribute of the TableTitle entity.

        :param is_floating: Whether the title is floating (not in-table) or not (in-table).
        :type is_floating: bool
        """

        self._is_floating = is_floating

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ):
        text, words = linearize_children(self.words, config=config)
        return text, words



"""
Represents a single :class:`Signature` Entity within the :class:`Document`. 
The Textract API response returns signatures as SIGNATURE BlockTypes.

This class contains the associated metadata with the :class:`Signature` entity including the entity ID, 
bounding box information, page number, Page ID and confidence of detection.
"""

import logging
from typing import List
import uuid
from textractor.data.text_linearization_config import TextLinearizationConfig

from textractor.entities.bbox import BoundingBox
from textractor.entities.document_entity import DocumentEntity
from textractor.entities.line import Line
from textractor.entities.word import Word


class Signature(DocumentEntity):
    """
    To create a new :class:`Signature` object we need the following:

    :param entity_id: Unique identifier of the signature entity.
    :type entity_id: str
    :param bbox: Bounding box of the signature entity.
    :type bbox: BoundingBox
    :param words: List of the Word entities present in the signature
    :type words: list, optional
    :param confidence: confidence with which the entity was detected.
    :type confidence: float, optional
    """

    def __init__(
        self,
        entity_id: str,
        bbox: BoundingBox,
        confidence: float = 0,
    ):
        super().__init__(entity_id, bbox)
        self._confidence = confidence / 100
        self._page = None
        self._page_id = None

    @property
    def page(self):
        """
        :return: Returns the page number of the page the :class:`Signature` entity is present in.
        :rtype: int
        """
        return self._page

    @property
    def words(self):
        """
        :return: Returns an empty list
        :rtype: list
        """
        return []

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the :class:`Signature` entity.

        :param page_num: Page number where the :class:`Signature` entity exists.
        :type page_num: int
        """
        self._page = page_num

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """
        return self._page_id

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the :class:`Signature` entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """
        self._page_id = page_id

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ):
        w = Word(
            entity_id=str(uuid.uuid4()), bbox=self.bbox, text=config.signature_token
        )
        w.line = Line(entity_id=str(uuid.uuid4()), bbox=self.bbox, words=[w])
        return config.signature_token, [w]



"""
The :class:`EntityList` is an extension of list type with custom functions to print document entities \
in a well formatted manner and visualize on top of the document page with their BoundingBox information. 

The two main functions within this class are :code:`pretty_print()` and :code:`visualize()`.
Use :code:`pretty_print()` to get a string formatted output of your custom list of entities.
Use :code:`visualize()` to get the bounding box visualization of the entities on the document page images.
"""

import os
import csv
import logging
from enum import Enum
from io import StringIO
from tabulate import tabulate
from typing import List, Optional, TypeVar, Generic, Any
from collections import defaultdict
from textractor.utils.geometry_util import get_indices
from PIL import Image, ImageDraw, ImageColor, ImageFont

from textractor.data.constants import (
    TextractType,
    TableFormat,
    AnalyzeExpenseLineItemFields,
    AnalyzeExpenseFields,
)
from textractor.exceptions import EntityListCreationError, NoImageException
from textractor.entities.linearizable import Linearizable
from textractor.data.text_linearization_config import TextLinearizationConfig

logger = logging.getLogger(__name__)

present_path = os.path.abspath(os.path.dirname(__file__))

T = TypeVar("T")


class EntityList(list, Generic[T], Linearizable):
    """
    Creates a list type object, initially empty but extended with the list passed in objs.

    :param objs: Custom list of objects that can be visualized with this class.
    :type objs: list
    """

    def __init__(self, objs=None):
        super().__init__()

        if objs is None:
            objs = []
        elif not isinstance(objs, list):
            objs = [objs]

        self.extend(objs)

    def visualize(
        self,
        with_text: bool = True,
        with_words: bool = True,
        with_confidence: bool = False,
        font_size_ratio: float = 0.5,
    ) -> List:
        """
        Returns list of PIL Images with bounding boxes drawn around the entities in the list.

        :param with_text: Flag to print the OCR output of Textract on top of the text bounding box.
        :type with_text: bool
        :param with_confidence: Flag to print the confidence of prediction on top of the entity bounding box.
        :type with_confidence: bool

        :return: Returns list of PIL Images with bounding boxes drawn around the entities in the list.
        :rtype: list
        """
        # FIXME: This is inelegant
        if len(self) > 0 and any(
            [ent.__class__.__name__ == "Document" for ent in self]
        ):
            return EntityList(self[0].pages).visualize(
                with_text=with_text,
                with_words=with_words,
                with_confidence=with_confidence,
                font_size_ratio=font_size_ratio,
            )
        elif len(self) > 0 and any([ent.__class__.__name__ == "Page" for ent in self]):
            new_entity_list = []
            for entity in self:
                if not with_words and (
                    entity.__class__.__name__ == "Word"
                    or entity.__class__.__name__ == "Line"
                ):
                    continue
                if entity.__class__.__name__ == "Page":
                    if with_words:
                        new_entity_list.extend(entity.words)
                        new_entity_list.extend(entity.lines)
                    new_entity_list.extend(entity.tables)
                    new_entity_list.extend(entity.key_values)
                    new_entity_list.extend(entity.checkboxes)
                    new_entity_list.extend(entity.layouts)
                    for expense_document in entity.expense_documents:
                        new_entity_list = self._add_expense_document_to_list(
                            new_entity_list, expense_document
                        )
                elif entity.__class__.__name__ == "ExpenseDocument":
                    self._add_expense_document_to_list(new_entity_list, entity)
                else:
                    new_entity_list.append(entity)
            return EntityList(list(dict.fromkeys(new_entity_list).keys())).visualize(
                with_text=with_text,
                with_words=with_words,
                with_confidence=with_confidence,
                font_size_ratio=font_size_ratio,
            )
        elif len(self) > 0 and self[0].bbox.spatial_object.image is None:
            raise NoImageException(
                "Image was not saved during the Textract API call. Set save_image=True when calling the Textractor methods to use the visualize() method."
            )

        visualized_images = {}
        entities_pagewise = defaultdict(list)
        for obj in self:
            entities_pagewise[obj.page].append(obj)
            if obj.page is None:
                print(obj.__class__.__name__)
            try:
                if with_words:
                    entities_pagewise[obj.page].extend(obj.words)
            # FIXME: There should be a way to recurse through all entities
            except AttributeError:
                pass

        for page in list(entities_pagewise.keys()):
            # Deduplication
            entities_pagewise[page] = list(dict.fromkeys(entities_pagewise[page]).keys())

        for page in entities_pagewise.keys():
            visualized_images[page] = _draw_bbox(
                entities_pagewise[page],
                with_text,
                with_confidence,
                font_size_ratio,
            )

        images = [image.convert("RGB") for image in visualized_images.values()]
        images = images if len(images) != 1 else images[0]
        return images

    def _add_expense_document_to_list(self, entity_list, expense_document):
        entity_list.append(expense_document)
        for field in expense_document.summary_fields_list:
            entity_list.append(field)
        for line_item_group in expense_document.line_items_groups:
            entity_list.append(line_item_group)
            for row in line_item_group.rows:
                entity_list.append(row)
                for expense in row.expenses:
                    if (
                        expense.type.text
                        != AnalyzeExpenseLineItemFields.EXPENSE_ROW.name
                    ):
                        entity_list.append(expense)
        return entity_list

    def pretty_print(
        self,
        table_format: TableFormat = TableFormat.GITHUB,
        with_confidence: bool = False,
        with_geo: bool = False,
        with_page_number: bool = False,
        trim: bool = False,
    ) -> str:
        """
        Returns a formatted string output for each of the entities in the list according to its entity type.

        :param table_format: Choose one of the defined TableFormat types to decorate the table output string. This is a predefined set of choices by the PyPI tabulate package. It is used only if there are KeyValues or Tables in the list of textractor.entities.
        :type table_format: TableFormat
        :param with_confidence: Flag to add the confidence of prediction to the entity string. default= False.
        :type with_confidence: bool
        :param with_geo: Flag to add the bounding box information to the entity string. default= False.
        :type with_geo: bool
        :param with_page_number: Flag to add the page number to the entity string. default= False.
        :type with_page_number: bool
        :param trim: Flag to trim text in the entity string. default= False.
        :type trim: bool

        :return: Returns a formatted string output for each of the entities in the list according to its entity type.
        :rtype: str
        """

        result_value = ""
        result_value += self._get_text_string(
            with_page_number=with_page_number,
            with_confidence=with_confidence,
            trim=trim,
            textract_type=TextractType.WORDS,
        )
        result_value += self._get_text_string(
            with_page_number=with_page_number,
            with_confidence=with_confidence,
            trim=trim,
            textract_type=TextractType.LINES,
        )
        result_value += self._get_forms_string(
            table_format=table_format,
            with_confidence=with_confidence,
            with_geo=with_geo,
            trim=trim,
            textract_type=TextractType.KEY_VALUE_SET,
        )
        result_value += self._get_forms_string(
            table_format=table_format,
            with_confidence=with_confidence,
            with_geo=with_geo,
            trim=trim,
            textract_type=TextractType.SELECTION_ELEMENT,
        )
        result_value += self._get_tables_string(
            table_format=table_format,
            with_confidence=with_confidence,
            with_geo=with_geo,
            trim=trim,
        )
        result_value += self._get_queries_string()
        result_value += self._get_expense_documents_string()
        result_value += self._get_id_documents_string()
        return result_value

    def _get_text_string(
        self,
        with_page_number=False,
        with_confidence=False,
        trim=False,
        textract_type=TextractType.WORDS,
    ):
        """
        Returns a formatted string output for the entity type stated in the textract_type param. This function is
        specific to TextractType.WORDS and TextractType.LINES.

        :param with_page_number: Flag to add the page number to the entity string. default= False.
        :type with_page_number: bool
        :param with_confidence: Flag to add the confidence of prediction to the entity string. default= False.
        :type with_confidence: bool
        :param trim: Flag to trim text in the entity string. default= False.
        :type trim: bool
        :param textract_type: TextractType.WORDS / TextractType.LINES
        :type textract_type: TextractType

        :return: Returns a formatted string output for the entity type stated in the textract_type param.
        :rtype: str
        """

        result_value = ""

        if textract_type == TextractType.WORDS:
            objects = sorted(
                [obj for obj in self if obj.__class__.__name__ == "Word"],
                key=lambda x: x.page,
            )
        else:
            objects = sorted(
                [obj for obj in self if obj.__class__.__name__ == "Line"],
                key=lambda x: x.page,
            )

        current_page = -1

        for word in objects:
            if with_page_number and word.page != current_page:
                result_value += f"--------- page number: {word.page} - page ID: {word.page_id} --------------{os.linesep}"
                current_page = word.page
            if trim:
                result_value += f"{word.text.strip()}"
            else:
                result_value += f"{word.text}"

            if with_confidence:
                result_value += f", {word.confidence}"
            result_value += os.linesep

        return result_value

    def _get_forms_string(
        self,
        table_format: TableFormat = TableFormat.GITHUB,
        with_confidence: bool = False,
        with_geo: bool = False,
        trim: bool = False,
        textract_type=TextractType.KEY_VALUE_SET,
    ) -> str:
        """
        Returns a formatted string output for the entity type stated in the textract_type param. This function is
        specific to TextractType.KEY_VALUE_SET and TextractType.SELECTION_ELEMENT.

        :param table_format: Choose one of the defined TableFormat types to decorate the table output string.
                             This is a predefined set of choices by the PyPI tabulate package.
        :type table_format: TableFormat
        :param with_confidence: Flag to add the confidence of prediction to the entity string. default= False.
        :type with_confidence: bool
        :param with_geo: Flag to add the bounding box information to the entity string. default= False.
        :type with_geo: bool
        :param trim: Flag to trim text in the entity string. default= False.
        :type trim: bool
        :param textract_type: TextractType.KEY_VALUE_SET / TextractType.SELECTION_ELEMENT
        :type textract_type: TextractType

        :return: Returns a formatted string output for the entity type stated in the textract_type param.
        :rtype: str
        """

        logger.debug(f"table_format: {table_format}")
        result_value = ""

        if textract_type == TextractType.KEY_VALUE_SET:
            key_value_objects = [
                obj
                for obj in self
                if obj.__class__.__name__ == "KeyValue" and not obj.contains_checkbox
            ]
        else:
            key_value_objects = [
                obj
                for obj in self
                if obj.__class__.__name__ == "KeyValue" and obj.contains_checkbox
            ]

        kv_dict = {obj.page: [] for obj in key_value_objects}

        for obj in key_value_objects:
            kv_dict[obj.page].append(obj)

        if not table_format == TableFormat.CSV:
            for page in kv_dict.keys():
                forms_list = _convert_form_to_list(
                    kv_dict[page],
                    with_confidence=with_confidence,
                    with_geo=with_geo,
                    trim=trim,
                    textract_type=textract_type,
                )
                result_value += (
                    tabulate(forms_list, tablefmt=table_format.name.lower())
                    + os.linesep
                    + os.linesep
                )

        if table_format == TableFormat.CSV:
            logger.debug(f"pretty print - csv")
            csv_output = StringIO()
            csv_writer = csv.writer(
                csv_output, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
            )
            for page in kv_dict.keys():
                forms_list = _convert_form_to_list(
                    kv_dict[page],
                    with_confidence=with_confidence,
                    with_geo=with_geo,
                    trim=trim,
                    textract_type=textract_type,
                )
                csv_writer.writerows(forms_list)
            csv_writer.writerow([])
            result_value = csv_output.getvalue()
        return result_value

    def _get_tables_string(
        self,
        table_format: TableFormat = TableFormat.GITHUB,
        with_confidence: bool = False,
        with_geo: bool = False,
        trim: bool = False,
    ) -> str:
        """
        Returns a formatted string output for the Table entity type.

        :param table_format: Choose one of the defined TableFormat types to decorate the table output string.
                             This is a predefined set of choices by the PyPI tabulate package.
        :type table_format: TableFormat
        :param with_confidence: Flag to add the confidence of prediction to the entity string. default= False.
        :type with_confidence: bool
        :param with_geo: Flag to add the bounding box information to the entity string. default= False.
        :type with_geo: bool
        :param trim: Flag to trim text in the entity string. default= False.
        :type trim: bool

        :return: Returns a formatted string output for the Table entity type.
        :rtype: str
        """

        logger.debug(f"table_format: {table_format}")

        tables = {}
        for obj in self:
            if obj.__class__.__name__ == "Table":
                tables[obj.id] = obj
            elif obj.__class__.__name__ == "TableCell":
                if obj.table_id in tables.keys():
                    tables[obj.table_id].append(obj)
                else:
                    tables[obj.table_id] = [obj]

        result_value = ""
        if not table_format == TableFormat.CSV:
            for table_id in tables.keys():
                table_type = (
                    TextractType.TABLES
                    if tables[table_id].__class__.__name__ == "Table"
                    else TextractType.TABLE_CELL
                )

                table_list = _convert_table_to_list(
                    tables[table_id],
                    with_confidence=with_confidence,
                    with_geo=with_geo,
                    trim=trim,
                    textract_type=table_type,
                )
                result_value += (
                    tabulate(table_list, tablefmt=table_format.name.lower())
                    + os.linesep
                    + os.linesep
                )

        if table_format == TableFormat.CSV:
            logger.debug(f"pretty print - csv")
            for table_id in tables.keys():
                table_type = (
                    TextractType.TABLES
                    if tables[table_id].__class__.__name__ == "Table"
                    else TextractType.TABLE_CELL
                )
                csv_output = StringIO()
                csv_writer = csv.writer(
                    csv_output, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
                )

                table_list = _convert_table_to_list(
                    tables[table_id],
                    with_confidence=with_confidence,
                    with_geo=with_geo,
                    trim=trim,
                    textract_type=table_type,
                )
                csv_writer.writerows(table_list)
                csv_writer.writerow([])
                result_value += csv_output.getvalue()
        return result_value

    def _get_queries_string(self):
        result_value = ""
        queries = [obj for obj in self if obj.__class__.__name__ == "Query"]

        for query in queries:
            if query.result is not None:
                result_value += f"{query.query} => {query.result.answer}{os.linesep}"
            else:
                result_value += f"{query.query} => {os.linesep}"

        return result_value

    def _get_expense_documents_string(self):
        result_value = ""
        expense_documents = [
            obj for obj in self if obj.__class__.__name__ == "ExpenseDocument"
        ]

        for i, expense_document in enumerate(expense_documents):
            result_value += f"Expense Document {i+1}:{os.linesep}"
            result_value += f"### Summary Fields:{os.linesep}"
            result_value += f"{expense_document.summary_fields}{os.linesep}"
            result_value += f"### Line Item Groups: {os.linesep}"
            for line_item_group in expense_document.line_items_groups:
                result_value += f"{line_item_group}{os.linesep}"

        return result_value

    def _get_id_documents_string(self):
        result_value = ""
        id_documents = [
            obj for obj in self if obj.__class__.__name__ == "IdentityDocument"
        ]

        for id_document in id_documents:
            result_value += f"{id_document}{os.linesep}"

        return result_value

    def __add__(self, list2):
        return EntityList([*self, *list2])

    def get_text_and_words(self, config: TextLinearizationConfig = TextLinearizationConfig()):
        texts, words = [], []
        separator = (
            config.same_paragraph_separator
            if all([entity.__class__.__name__ == "Word" for entity in self]) else
            config.layout_element_separator
        )
        for entity in self:
            entity_text, entity_words = entity.get_text_and_words(config)
            texts.append(entity_text)
            words.extend(entity_words)
        return separator.join(texts), words

def _convert_form_to_list(
    form_objects,
    with_confidence: bool = False,
    with_geo: bool = False,
    trim: bool = False,
    textract_type=TextractType.KEY_VALUE_SET,
) -> List:
    """
    Converts KeyValue objects (KEY_VALUE_SET in JSON) to row-wise list format to pretty_print using the
    PyPI tabulate package.

    :param form_objects: KeyValue instances to be formatted into strings
    :type form_objects: KeyValue
    :param with_confidence: Flag to add the confidence of prediction to the entity string. default= False.
    :type with_confidence: bool
    :param with_geo: Flag to add the bounding box information to the entity string. default= False.
    :type with_geo: bool
    :param trim: Flag to trim text in the entity string. default= False.
    :type trim: bool
    :param textract_type: TextractType.KEY_VALUE_SET / TextractType.SELECTION_ELEMENT
    :type textract_type: TextractType

    :return: Returns a list of lists, each inner list containing a key-value pair.
    :rtype: List[List[str]]
    """

    rows_list = list()
    rows_list.append(["Key", "Value"])
    for field in form_objects:
        t_key = ""
        t_value = ""
        if field.key:
            text = " ".join([word.text for word in field.key])
            if trim:
                t_key = text.strip()
            else:
                t_key = text
            if with_geo:
                t_key += " {" + field.bbox.__repr__() + "} "
            if with_confidence:
                t_key += f" ({field.key.confidence:.1f})"
        if field.value:
            text = (
                field.value.words
                if textract_type == TextractType.SELECTION_ELEMENT
                else " ".join([word.text for word in field.value.words])
            )
            if trim:
                t_value = text.strip()
            else:
                t_value = text
            if with_geo:
                t_value += " {" + field.value.bbox.__repr__() + "} "
            if with_confidence:
                t_value += f" ({field.value.confidence:.1f})"
        rows_list.append([t_key, t_value])
    return rows_list


def _convert_table_to_list(
    table_object,
    with_confidence: bool = False,
    with_geo: bool = False,
    trim: bool = False,
    textract_type=TextractType.TABLES,
) -> List:
    """
    Converts Table objects (TABLE in JSON) to row-wise list format to pretty_print using the
    PyPI tabulate package.

    :param table_object: Table instance to be formatted into strings
    :type table_object: Table
    :param with_confidence: Flag to add the confidence of prediction to the entity string. default= False.
    :type with_confidence: bool
    :param with_geo: Flag to add the bounding box information to the entity string. default= False.
    :type with_geo: bool
    :param trim: Flag to trim text in the entity string. default= False.
    :type trim: bool
    :param textract_type: TextractType.TABLES / TextractType.TABLE_CELL
    :type textract_type: TextractType

    :return: Returns a list of lists, each inner list containing a row of table data.
    :rtype: List[List]
    """
    if textract_type == TextractType.TABLES:
        rowwise_table = table_object._get_table_cells()
    else:
        rowwise_table = {cell.row_index: [] for cell in table_object}
        for cell in table_object:
            rowwise_table[cell.row_index].append(cell)
    table_rows = []
    for row in rowwise_table.keys():
        row_data = []
        for cell in rowwise_table[row]:
            text = cell.__repr__().split(">")[-1][1:]
            if trim:
                t_key = text.strip()
            else:
                t_key = text
            if with_geo:
                t_key += " {" + cell.bbox.__repr__() + "} "
            if with_confidence:
                t_key += f" ({cell.confidence:.1f})"
            row_data.append(t_key)
        table_rows.append(row_data)
    return table_rows


def _draw_bbox(
    entities: List[Any],
    with_text: bool = False,
    with_confidence: bool = False,
    font_size_ratio: float = 0.5,
):
    """
    Function to draw bounding boxes on all objects in entities present in a particular page.

    :param entities: List of entities to be visualized on top of the document page
    :type entities: list, required
    :param with_text: Flag to indicate if text is to be printed on top of the bounding box
    :type with_text: bool, optional
    :param with_word_text_only: Flag to print only the word-level OCR output of Textract on top of the text bounding box.
    :type with_word_text_only: bool
    :param with_confidence: Flag to print the confidence of prediction on top of the entity bounding box.
    :type with_confidence: bool
    :param with_word_confidence_only: Flag to print only the word-level confidence of Textract OCR.
    :type with_word_confidence_only: bool

    :return: Returns PIL.Image with bounding boxes drawn for the entities passed to the function
    :rtype: PIL.Image
    """
    image = entities[0].bbox.spatial_object.image
    if image is None:
        for e in entities:
            if e.bbox.spatial_object.image is not None:
                image = e.bbox.spatial_object.image
                break
        else:
            raise Exception("Could not find an entity with an associated image!")
    image = image.convert("RGBA")
    overlay = Image.new("RGBA", image.size, (255, 255, 255, 0))
    drw = ImageDraw.Draw(overlay, "RGBA")

    text_locations = {}

    # First drawing tables
    for entity in entities:
        width, height = image.size
        if entity.__class__.__name__ == "Table":
            overlayer_data = _get_overlayer_data(entity, width, height)
            drw.rectangle(
                xy=overlayer_data["coords"], outline=overlayer_data["color"], width=2
            )
            if entity.title:
                drw.rectangle(
                    (
                        int(entity.title.bbox.x * width),
                        int(entity.title.bbox.y * height),
                        int((entity.title.bbox.x + entity.title.bbox.width) * width),
                        int((entity.title.bbox.y + entity.title.bbox.height) * height),
                    ),
                    outline=overlayer_data["color"],
                    fill=ImageColor.getrgb("red") + (120,),
                    width=2,
                )

            for footer in entity.footers:
                drw.rectangle(
                    (
                        int(footer.bbox.x * width),
                        int(footer.bbox.y * height),
                        int((footer.bbox.x + footer.bbox.width) * width),
                        int((footer.bbox.y + footer.bbox.height) * height),
                    ),
                    outline=overlayer_data["color"],
                    fill=ImageColor.getrgb("cyan") + (120,),
                    width=2,
                )

            processed_cells = set()
            for cell in entity.table_cells:
                if cell.id in processed_cells:
                    continue
                if cell.siblings:
                    for c in cell.siblings:
                        processed_cells.add(c.id)
                    min_x, min_y, max_x, max_y = list(
                        zip(
                            *[
                                (
                                    c.bbox.x,
                                    c.bbox.y,
                                    c.bbox.x + c.bbox.width,
                                    c.bbox.y + c.bbox.height,
                                )
                                for c in cell.siblings + [cell]
                            ]
                        )
                    )
                    min_x, min_y, max_x, max_y = (
                        min(min_x),
                        min(min_y),
                        max(max_x),
                        max(max_y),
                    )
                else:
                    processed_cells.add(cell.id)
                    min_x, min_y, max_x, max_y = (
                        cell.bbox.x,
                        cell.bbox.y,
                        cell.bbox.x + cell.bbox.width,
                        cell.bbox.y + cell.bbox.height,
                    )

                fill_color = None
                if cell.is_column_header:
                    fill_color = ImageColor.getrgb("blue") + (120,)
                if cell.is_title:
                    fill_color = ImageColor.getrgb("red") + (120,)
                if cell.is_footer:
                    fill_color = ImageColor.getrgb("cyan") + (120,)
                if cell.is_summary:
                    fill_color = ImageColor.getrgb("yellow") + (120,)
                if cell.is_section_title:
                    fill_color = ImageColor.getrgb("green") + (120,)

                drw.rectangle(
                    (
                        int(min_x * width),
                        int(min_y * height),
                        int(max_x * width),
                        int(max_y * height),
                    ),
                    outline=overlayer_data["color"],
                    fill=fill_color,
                    width=2,
                )
                for checkbox in cell.checkboxes:
                    drw.rectangle(
                        (
                            int(checkbox.bbox.x * width),
                            int(checkbox.bbox.y * height),
                            int((checkbox.bbox.x + checkbox.bbox.width) * width),
                            int((checkbox.bbox.y + checkbox.bbox.height) * height),
                        ),
                        outline=ImageColor.getrgb("lightgreen")
                        if checkbox.is_selected()
                        else ImageColor.getrgb("indianred"),
                    )
    # Second drawing bounding boxes
    for entity in entities:
        if entity.__class__.__name__ == "Query":
            overlayer_data = _get_overlayer_data(entity.result, width, height)
            drw.rectangle(
                xy=overlayer_data["coords"], outline=overlayer_data["color"], width=2
            )
        elif entity.__class__.__name__ == "TableTitle":
            overlayer_data = _get_overlayer_data(entity.result, width, height)
            drw.rectangle(
                xy=overlayer_data["coords"], outline=overlayer_data["color"], width=2
            )
        elif entity.__class__.__name__ == "TableFooter":
            overlayer_data = _get_overlayer_data(entity.result, width, height)
            drw.rectangle(
                xy=overlayer_data["coords"], outline=overlayer_data["color"], width=2
            )
        elif entity.__class__.__name__ == "ExpenseField":
            overlayer_data = _get_overlayer_data(entity, width, height)
            drw.rectangle(
                xy=overlayer_data["coords_value"],
                outline=overlayer_data["color_value"],
                width=2,
            )
            if entity.key is not None:
                b1 = entity.key.bbox
                b2 = entity.value.bbox
                drw.rectangle(
                    xy=overlayer_data["coords_key"],
                    outline=overlayer_data["color_key"],
                    width=2,
                )
                drw.line(
                    [
                        (
                            (b1.x + b1.width / 2) * width,
                            (b1.y + b1.height / 2) * height,
                        ),
                        (
                            (b2.x + b2.width / 2) * width,
                            (b2.y + b2.height / 2) * height,
                        ),
                    ],
                    fill=overlayer_data["color_key"],
                    width=2,
                )
        elif entity.__class__.__name__ == "ExpenseDocument":
            overlayer_data = _get_overlayer_data(entity, width, height)
            drw.rectangle(
                xy=overlayer_data["coords"], outline=overlayer_data["color"], width=2
            )
            for coord, text in zip(
                overlayer_data["coords_list"], overlayer_data["coords_list"]
            ):
                drw.rectangle(
                    xy=coord, outline=overlayer_data["color_expense_group"], width=2
                )
        elif entity.__class__.__name__.startswith("Layout"):
            overlayer_data = _get_overlayer_data(entity, width, height)
            drw.rectangle(
                xy=overlayer_data["coords"], outline=overlayer_data["color"], width=4
            )
        else:
            overlayer_data = _get_overlayer_data(entity, width, height)
            drw.rectangle(
                xy=overlayer_data["coords"], outline=overlayer_data["color"], width=2
            )
            if entity.__class__.__name__ == "KeyValue":
                drw.rectangle(
                    xy=overlayer_data["value_bbox"],
                    outline=overlayer_data["color_value"],
                    width=2,
                )
                b1 = overlayer_data["value_bbox"]
                b2 = overlayer_data["coords"]
                drw.line(
                    [
                        ((b1[0] + b1[2]) / 2, (b1[1] + b1[3]) / 2),
                        ((b2[0] + b2[2]) / 2, (b2[1] + b2[3]) / 2),
                    ],
                    fill=overlayer_data["color_value"],
                    width=1,
                )

    # Second drawing, text
    if with_text:
        for entity in entities:
            if entity.__class__.__name__ == "Word":
                width, height = image.size
                overlayer_data = _get_overlayer_data(entity, width, height)

                final_txt = ""
                bbox_height = overlayer_data["coords"][3] - overlayer_data["coords"][1]
                text_height = int(bbox_height * font_size_ratio)
                fnt = ImageFont.truetype(
                    os.path.join(present_path, "arial.ttf"), text_height
                )

                final_txt += overlayer_data["text"]

                if with_confidence:
                    final_txt += " (" + str(overlayer_data["confidence"])[:4] + ")"

                drw.text(
                    (
                        overlayer_data["coords"][0],
                        overlayer_data["coords"][1] - text_height,
                    ),
                    final_txt,
                    font=fnt,
                    fill=overlayer_data["text_color"],
                )

            elif entity.__class__.__name__ == "KeyValue":
                width, height = image.size
                overlayer_data = _get_overlayer_data(entity, width, height)

                # Key Text
                final_txt = ""
                bbox_height = overlayer_data["coords"][3] - overlayer_data["coords"][1]
                text_height = min(
                    int(0.03 * height), int(bbox_height * font_size_ratio)
                )
                fnt = ImageFont.truetype(
                    os.path.join(present_path, "arial.ttf"), text_height
                )

                final_txt += overlayer_data["text"]
                if with_confidence:
                    final_txt += " (" + str(overlayer_data["confidence"])[:4] + ")"

                drw.text(
                    (
                        overlayer_data["coords"][0],
                        overlayer_data["coords"][3] + 1,
                    ),
                    final_txt,
                    font=fnt,
                    fill=overlayer_data["color"],
                )

                # Value Text
                final_txt = overlayer_data["value_text"]

                bbox_height = (
                    overlayer_data["value_bbox"][3] - overlayer_data["value_bbox"][1]
                )
                text_height = min(
                    int(0.01 * height), int(bbox_height * font_size_ratio)
                )
                fnt = ImageFont.truetype(
                    os.path.join(present_path, "arial.ttf"), text_height
                )

                if with_confidence:
                    final_txt += " (" + str(overlayer_data["value_conf"])[:4] + ")"

                drw.text(
                    (
                        overlayer_data["value_bbox"][0],
                        overlayer_data["value_bbox"][3] + 1,
                    ),
                    final_txt,
                    font=fnt,
                    fill=overlayer_data["color_value"],
                )
            elif entity.__class__.__name__ == "ExpenseField":
                width, height = image.size
                overlayer_data = _get_overlayer_data(entity, width, height)

                final_txt = overlayer_data["text"]
                text_height = int(0.018 * height * font_size_ratio)
                fnt = ImageFont.truetype(
                    os.path.join(present_path, "arial.ttf"), text_height
                )
                if entity.key is not None:
                    x = overlayer_data["coords_key"][0] + 0.3 * (
                        overlayer_data["coords_key"][2]
                        - overlayer_data["coords_key"][0]
                    )
                    y = overlayer_data["coords_key"][1] - text_height - 1
                else:
                    x = int(
                        overlayer_data["coords"][0]
                        + 0.3
                        * (overlayer_data["coords"][2] - overlayer_data["coords"][0])
                    )
                    y = overlayer_data["coords"][1] - text_height - 1
                while (x, y) in text_locations and text_locations[(x, y)] != final_txt:
                    y = y - text_height - 1
                text_locations[(x, y)] = final_txt
                drw.text(
                    (x, y),
                    final_txt,
                    font=fnt,
                    fill=overlayer_data["text_color"],
                )
            elif entity.__class__.__name__ == "ExpenseDocument":
                width, height = image.size
                text_height = int(0.018 * height * font_size_ratio)
                fnt = ImageFont.truetype(
                    os.path.join(present_path, "arial.ttf"), text_height
                )
                overlayer_data = _get_overlayer_data(entity, width, height)
                for coord, text in zip(
                    overlayer_data["coords_list"], overlayer_data["text_list"]
                ):
                    drw.text(
                        (coord[0], coord[3]),
                        text,
                        font=fnt,
                        fill=overlayer_data["color_expense_group"],
                    )

            elif entity.__class__.__name__ == "Query":
                if entity.result is None:
                    continue

                width, height = image.size
                overlayer_data = _get_overlayer_data(entity.result, width, height)

                final_txt = entity.query + " " + overlayer_data["text"]

                bbox_height = overlayer_data["coords"][3] - overlayer_data["coords"][1]
                text_height = int(bbox_height * font_size_ratio)
                fnt = ImageFont.truetype(
                    os.path.join(present_path, "arial.ttf"), text_height
                )

                if with_confidence:
                    final_txt += " (" + str(entity.result.confidence)[:4] + ")"

                drw.text(
                    (
                        overlayer_data["coords"][0],
                        overlayer_data["coords"][1] - text_height,
                    ),
                    final_txt,
                    font=fnt,
                    fill=overlayer_data["text_color"],
                )

            elif entity.__class__.__name__.startswith("Layout"):
                width, height = image.size
                overlayer_data = _get_overlayer_data(entity, width, height)

                final_txt = ""
                bbox_height = overlayer_data["coords"][3] - overlayer_data["coords"][1]
                text_height = min(20, int(bbox_height * font_size_ratio))
                fnt = ImageFont.truetype(
                    os.path.join(present_path, "arial.ttf"), text_height
                )

                final_txt += overlayer_data["text"]

                if with_confidence:
                    final_txt += " (" + str(overlayer_data["confidence"])[:4] + ")"

                drw.text(
                    (
                        overlayer_data["coords"][0],
                        overlayer_data["coords"][1] - text_height,
                    ),
                    final_txt,
                    font=fnt,
                    fill=overlayer_data["text_color"],
                )
    del drw
    image = Image.alpha_composite(image, overlay)
    return image


def _get_overlayer_data(entity: Any, width: float, height: float) -> dict:
    """
    Returns a dictionary with all the necessary details to draw a bounding box for an entity depending on the information
    present in it. This includes the bounding box coordinates, color of bounding box, confidence of detection and OCR text.

    :param entity: DocumentEntity object for which the data needs to be created
    :type entity: DocumentEntity
    :param width: width of the Page object the entity belongs to
    :type width: float, required
    :param height: height of the Page object the entity belongs to
    :type height: float, required

    :return: Dictionary containing all the information to draw the bounding box for a DocumentEntity.
    :rtype: dict
    """
    data = {}
    bbox = entity.bbox
    x, y, w, h = (
        bbox.x * width,
        bbox.y * height,
        bbox.width * width,
        bbox.height * height,
    )
    data["coords"] = [x, y, x + w, y + h]
    data["confidence"] = (
        entity.confidence
        if entity.__class__.__name__
        not in [
            "Table",
            "ExpenseField",
            "ExpenseDocument",
            "LineItemRow",
            "LineItemGroup",
        ]
        else ""
    )
    data["color"] = (0, 0, 0)
    data["text_color"] = (0, 0, 0)

    if entity.__class__.__name__ == "Word":
        data["text"] = entity.text
        data["color"] = ImageColor.getrgb("blue")

    elif entity.__class__.__name__ == "Line":
        data["text"] = entity.text
        data["color"] = ImageColor.getrgb("lightgrey")
        data["coords"] = [x - 1, y - 1, x + w + 1, y + h + 1]
    elif entity.__class__.__name__ == "KeyValue":
        data["text"] = entity.key.__repr__()
        data["color"] = ImageColor.getrgb("brown")
        data["value_text"] = entity.value.__repr__()
        data["coords"] = [x - 2, y - 2, x + w + 2, y + h + 2]

        if entity.contains_checkbox and entity.children:
            value_bbox = entity.children[0].bbox
            data["value_conf"] = entity.children[0].confidence

        else:
            value_bbox = entity.value.bbox
            data["value_conf"] = entity.value.confidence

        data["color_value"] = ImageColor.getrgb("orange")
        x, y, w, h = (
            value_bbox.x * width - 2,
            value_bbox.y * height - 2,
            value_bbox.width * width + 2,
            value_bbox.height * height + 2,
        )
        data["value_bbox"] = [x, y, x + w, y + h]

    elif entity.__class__.__name__ == "Table":
        data["color"] = ImageColor.getrgb("green")
        data["text"] = ""
    elif entity.__class__.__name__ == "TableTitle":
        data["color"] = ImageColor.getrgb("green")
        data["text"] = ""
    elif entity.__class__.__name__ == "TableFooter":
        data["color"] = ImageColor.getrgb("green")
        data["text"] = ""

    elif entity.__class__.__name__ == "TableCell":
        data["color"] = ImageColor.getrgb("skyblue")
        data["text"] = entity.__repr__().split(">")[-1][1:]

    elif entity.__class__.__name__ == "QueryResult":
        data["color"] = ImageColor.getrgb("mediumturquoise")
        data["text"] = entity.answer
    elif entity.__class__.__name__ == "Signature":
        data["color"] = ImageColor.getrgb("coral")
    elif entity.__class__.__name__ == "ExpenseField":
        data["text"] = entity.type.text
        data["text_color"] = ImageColor.getrgb("brown")
        data["coords"] = [x - 5, y - 5, x + w + 5, y + h + 5]

        if entity.key:
            data["color_key"] = ImageColor.getrgb("brown")
            data["coords_key"] = (
                entity.key.bbox.x * width - 3,
                entity.key.bbox.y * height - 3,
                (entity.key.bbox.x + entity.key.bbox.width) * width + 3,
                ((entity.key.bbox.y + entity.key.bbox.height)) * height + 3,
            )
        data["color_value"] = ImageColor.getrgb("orange")
        data["coords_value"] = (
            entity.value.bbox.x * width - 3,
            entity.value.bbox.y * height - 3,
            (entity.value.bbox.x + entity.value.bbox.width) * width + 3,
            ((entity.value.bbox.y + entity.value.bbox.height)) * height + 3,
        )
    elif entity.__class__.__name__ == "Expense":
        data["text"] = entity.text
        data["coords"] = [x - 3, y - 3, x + w + 3, y + h + 3]
    elif entity.__class__.__name__ == "ExpenseDocument":
        data["color"] = ImageColor.getrgb("beige")
        data["coords_list"] = []
        data["text_list"] = []
        for group in entity.summary_groups:
            bboxes = entity.summary_groups.get_group_bboxes(group)
            for bbox in bboxes:
                data["coords_list"].append(
                    (
                        bbox.x * width - 5,
                        bbox.y * height - 5,
                        (bbox.x + bbox.width) * width + 3,
                        (bbox.y + bbox.height) * height + 3,
                    )
                )
                data["text_list"].append(group)
        data["color_expense_group"] = ImageColor.getrgb("coral")

    elif entity.__class__.__name__ == "LineItemGroup":
        data["color"] = ImageColor.getrgb("lightblue")
        data["coords"] = [x - 10, y - 10, x + w + 10, y + h + 10]

    elif entity.__class__.__name__ == "LineItemRow":
        data["color"] = ImageColor.getrgb("lightyellow")
        data["coords"] = [x - 7, y - 7, x + w + 7, y + h + 7]
    elif entity.__class__.__name__.startswith("Layout"):
        data["color"] = ImageColor.getrgb("teal")
        data["text"] = f"{entity.layout_type} - {entity.reading_order}"
    else:
        pass
    return data



"""
Represents a single :class:`TableCell:class:` object. The :class:`TableCell` objects contains information such as:

* The position info of the cell within the encompassing Table
* Properties such as merged-cells span
* A hierarchy of words contained within the TableCell (optional)
* Page information
* Confidence of entity detection.
"""

import os
from typing import List, Tuple

from textractor.entities.word import Word
from textractor.exceptions import InputError
from textractor.entities.bbox import BoundingBox
from textractor.visualizers.entitylist import EntityList
from textractor.entities.document_entity import DocumentEntity
from textractor.entities.selection_element import SelectionElement

from textractor.data.constants import (
    IS_COLUMN_HEAD,
    IS_MERGED_CELL,
    IS_SECTION_TITLE_CELL,
    IS_SUMMARY_CELL,
    IS_TITLE_CELL,
    IS_FOOTER_CELL,
)
from textractor.data.constants import TextTypes
from textractor.utils.text_utils import TextLinearizationConfig, linearize_children


class TableCell(DocumentEntity):
    """
    To create a new TableCell object we need the following:

    :param entity_id: Unique id of the TableCell object
    :param bbox: Bounding box of the entity
    :param row_index: Row index of position of cell within the table
    :param col_index: Column index of position of cell within the table
    :param row_span: How many merged cells does the cell spans horizontally (1 means no merged cells)
    :param col_span: How many merged cells does the cell spand vertically (1 means no merged cells)
    :param confidence: Confidence out of 100 with which the Cell was detected.
    :param is_column_header: Indicates if the cell is a column header
    :param is_title: Indicates if the cell is a table title
    :param is_footer: Indicates if the cell is a table footer
    :param is_summary: Indicates if the cell is a summary cell
    :param is_section_title: Indicates if the cell is a section title
    """

    def __init__(
        self,
        entity_id: str,
        bbox: BoundingBox,
        row_index: int,
        col_index: int,
        row_span: int,
        col_span: int,
        confidence: float = 0,
        is_column_header: bool = False,
        is_title: bool = False,
        is_footer: bool = False,
        is_summary: bool = False,
        is_section_title: bool = False,
    ):
        super().__init__(entity_id, bbox)
        self._row_index: int = int(row_index)
        self._col_index: int = int(col_index)
        self._row_span: int = int(row_span)
        self._col_span: int = int(col_span)
        self._words: List[Word] = []
        self._confidence = confidence / 100
        self._page = None
        self._page_id = None
        self._is_column_header = is_column_header
        self._is_title = is_title
        self._is_footer = is_footer
        self._is_summary = is_summary
        self._is_section_title = is_section_title
        # this gets populated when cells are added to a table using the `add_cells` method
        # or when cells are attributed to a table with table.cells = [TableCell]
        self._parent_table_id = None
        self.parent_cell_id = None
        self.siblings: List[TableCell] = []

    @property
    def is_column_header(self):
        return self._is_column_header

    @property
    def is_title(self):
        return self._is_title

    @property
    def is_footer(self):
        return self._is_footer

    @property
    def is_summary(self):
        return self._is_summary

    @property
    def is_section_title(self):
        return self._is_section_title

    @property
    def page(self):
        """
        :return: Returns the page number of the page the :class:`TableCell` entity is present in.
        :rtype: int
        """
        return self._page

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the :class:`TableCell` entity.

        :param page_num: Page number where the TableCell entity exists.
        :type page_num: int
        """
        self._page = page_num

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """
        return self._page_id

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the TableCell entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """
        self._page_id = page_id

    @property
    def row_index(self):
        """
        :return: Returns the row index of the cell in the :class:`Table`.
        :rtype: int
        """
        return self._row_index

    @row_index.setter
    def row_index(self, index: int):
        """
        Sets the row index of the cell in the Table.

        :param index: Row value of the cell in the Table.
        :type index: int
        """
        self._row_index = index

    @property
    def col_index(self):
        """
        :return: Returns the column index of the cell in the Table.
        :rtype: int
        """
        return self._col_index

    @col_index.setter
    def col_index(self, index: int):
        """
        Sets the column index of the cell in the :class:`Table`.

        :param index: Column value of the cell in the Table.
        :type index: int
        """
        self._col_index = index

    @property
    def row_span(self):
        """
        :return: Returns the row span of the cell in the :class:`Table`.
        :rtype: int
        """
        return self._row_span

    @property
    def col_span(self):
        """
        :return: Returns the column span of the cell in the :class:`Table`.
        :rtype: int
        """
        return self._col_span

    @property
    def words(self):
        """
        Returns all the Word objects present in the :class:`TableCell`.

        :return words: List of Word objects, each representing a word within the TableCell.
        :rtype: list
        """
        return EntityList(self.get_text_and_words()[1])

    @property
    def checkboxes(self):
        output = []
        for child in self._children:
            if isinstance(child, SelectionElement):
                output.append(child)
        return output

    @property
    def text(self) -> str:
        """Returns the text in the cell as one space-separated string

        :return: Text in the cell
        :rtype: str
        """
        return self.get_text()

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> Tuple[str, List]:
        """Returns the text in the cell as one space-separated string

        :return: Text in the cell
        :rtype: Tuple[str, List]
        """
        text, words = linearize_children(self.children, config=config, no_new_lines=True)
        return text, words

    @property
    def table_id(self):
        """
        :return: Returns the ID of the :class:`Table` the TableCell belongs to.
        :rtype: str
        """
        return self._parent_table_id

    @table_id.setter
    def table_id(self, id: str):
        """
        Sets the ID of the Table the TableCell belongs to.

        :param id: ID of the Table
        :type id: str
        """
        self._parent_table_id = id

    def _update_response_metadata(self, metadata: list):
        """
        Updates metadata of :class:`TableCell` to include information stating if cell is of a special type.

        :param metadata: List of string types that match different cell types
        :type metadata: List
        """
        self.metadata[IS_COLUMN_HEAD] = True if "COLUMN_HEADER" in metadata else False
        self.metadata[IS_MERGED_CELL] = True if "MERGED_CELL" in metadata else False
        self.metadata[IS_SECTION_TITLE_CELL] = (
            True if "SECTION_TITLE" in metadata else False
        )
        self.metadata[IS_SUMMARY_CELL] = True if "SUMMARY_CELL" in metadata else False
        self.metadata[IS_TITLE_CELL] = True if "FLOATING_TITLE" in metadata else False
        self.metadata[IS_FOOTER_CELL] = True if "FLOATING_FOOTER" in metadata else False

    def _get_merged_cell_range(self):
        """
        :return: Returns the first row, first column, last row and last column of the merged cell.
        :rtype: Tuple[float]
        """
        if self.metadata[IS_MERGED_CELL]:
            rows = set()
            cols = set()
            for cell in self.siblings:
                rows.add(cell.row_index)
                cols.add(cell.col_index)
            return min(rows), min(cols), max(rows), max(cols)
        else:
            return self.row_index, self.col_index, self.row_index, self.col_index

    def get_words_by_type(self, text_type: TextTypes = TextTypes.PRINTED) -> List[Word]:
        """
        Returns list of :class:`Word` entities that match the input text type.

        :param text_type: TextTypes.PRINTED or TextTypes.HANDWRITING
        :type text_type: TextTypes
        :return: Returns list of Word entities that match the input text type.
        :rtype: EntityList
        """
        if not isinstance(text_type, TextTypes):
            raise InputError(
                "text_type parameter should be of TextTypes type. Find input choices from textractor.data.constants"
            )

        return EntityList([word for word in self.words if word.text_type == text_type])

    def merge_direction(self):
        """
        :return:    Determines if the merged cell is a row or column merge.
                    Returns 0 if row merge, 1 if column merge and 2 if both and None if there is no merge.
        :rtype: int, str
        """
        if self.metadata[IS_MERGED_CELL]:
            rows = set()
            columns = set()
            for cell in self.siblings:
                rows.add(cell.row_index)
                columns.add(cell.col_index)
            if len(rows) > 1 and len(columns) > 1:
                return 2, "Row and Column"
            elif len(rows) > 1:
                return 1, "Column"
            elif len(columns) > 1:
                return 0, "Row"
            else:
                self.metadata[IS_MERGED_CELL] = False
        return None, "None"

    def __repr__(self):
        """
        :return: String representation of the TableCell entity.
        :rtype: str
        """
        if self.metadata.get(IS_MERGED_CELL, False):
            entities = {
                (cell.row_index, cell.col_index): sorted(
                    cell.words, key=lambda x: (x.bbox.x, x.bbox.y)
                )
                for cell in sorted(
                    self.siblings, key=lambda x: (x.row_index, x.col_index)
                )
            }

            rows = set(sorted([cell_key[0] for cell_key in entities.keys()]))
            cols = set(sorted([cell_key[1] for cell_key in entities.keys()]))

            entity_repr = []
            for row in rows:
                for col in cols:
                    entity_repr.append(
                        " ".join([entity.__repr__() for entity in entities[(row, col)]])
                    )
                    entity_repr.append(" ")
                entity_repr.append(os.linesep)
            entity_repr = "".join(entity_repr)

        else:
            entities = self.words
            entity_repr = " ".join([entity.__repr__() for entity in entities])

        entity_string = f"<Cell: ({self.row_index},{self.col_index}), Span: ({self.row_span}, {self.col_span}), Column Header: { self.is_column_header}, "
        entity_string += (
            f"MergedCell: {self.metadata.get(IS_MERGED_CELL, False)}>  " + entity_repr
        )
        return entity_string



"""Multiple feature calls within the Textractor tool require the use of input parameters from fixed choices. \
They've been defined as :code:`Enum` within the data package."""

from enum import Enum
from textractcaller.t_call import Textract_API

from textractor.exceptions import InputError

WORD = "WORD"
LINE = "LINE"
KEY_VALUE_SET = "KEY_VALUE_SET"
SELECTION_ELEMENT = "SELECTION_ELEMENT"
TABLE = "TABLE"
TABLE_TITLE = "TABLE_TITLE"
TABLE_FOOTER = "TABLE_FOOTER"
TABLE_SUMMARY = "TABLE_SUMMARY"
TABLE_SECTION_TITLE = "TABLE_SECTION_TITLE"
TABLE_COLUMN_HEADER = "COLUMN_HEADER"
TABLE_STRUCTURED = "STRUCTURED_TABLE"
TABLE_SEMI_STRUCTURED = "SEMI_STRUCTURED_TABLE"
CELL = "CELL"
PAGE = "PAGE"
MERGED_CELL = "MERGED_CELL"
QUERY = "QUERY"
QUERY_RESULT = "QUERY_RESULT"
SIGNATURE = "SIGNATURE"
LAYOUT = "LAYOUT"
LAYOUT_TEXT = "LAYOUT_TEXT"
LAYOUT_TITLE = "LAYOUT_TITLE"
LAYOUT_HEADER = "LAYOUT_HEADER"
LAYOUT_FOOTER = "LAYOUT_FOOTER"
LAYOUT_SECTION_HEADER = "LAYOUT_SECTION_HEADER"
LAYOUT_PAGE_NUMBER = "LAYOUT_PAGE_NUMBER"
LAYOUT_LIST = "LAYOUT_LIST"
LAYOUT_FIGURE = "LAYOUT_FIGURE"
LAYOUT_TABLE = "LAYOUT_TABLE"
LAYOUT_KEY_VALUE = "LAYOUT_KEY_VALUE"

# This is not a base layout type, but a fake type for the text conversion logic
# when an API is called without the AnalyzeLayout API, all the elements will be LAYOUT_ENTITY
LAYOUT_ENTITY = "LAYOUT_ENTITY"

# cell type attributes
IS_COLUMN_HEAD = "isColumnHead"
IS_SECTION_TITLE_CELL = "isSectionTitleCell"
IS_SUMMARY_CELL = "isSummaryCell"
IS_TITLE_CELL = "isTitleCell"
IS_FOOTER_CELL = "isFooterCell"
IS_MERGED_CELL = "isMergedCell"

# Text Types
HANDWRITING = "HANDWRITING"
PRINTED = "PRINTED"


class TableTypes(Enum):
    """Types of tables recognized by Textract APIs."""

    UNKNOWN = 0
    STRUCTURED = 1
    SEMI_STRUCTURED = 2


class Direction(Enum):
    """Directions available for search using DirectionalFinder"""

    ABOVE = 0
    BELOW = 1
    RIGHT = 2
    LEFT = 3


class TextTypes(Enum):
    """Textract recognizes TextType of all words in the document to fall into one of these 2 categories."""

    HANDWRITING = 0
    PRINTED = 1


# selection element attributes
SELECTED = "SELECTED"
NOT_SELECTED = "NOT_SELECTED"


class SelectionStatus(Enum):
    """These are the 2 categories defined for the SelectionStatus of a :class:`SelectionElement`."""

    SELECTED = 0
    NOT_SELECTED = 1


COLUMN_HEADER = "COLUMN_HEADER"
SECTION_TITLE = "SECTION_TITLE"
FLOATING_FOOTER = "FLOATING_FOOTER"
FLOATING_TITLE = "FLOATING_TITLE"
SUMMARY_CELL = "SUMMARY_CELL"


class CellTypes(Enum):
    """Special cells within the :class:`Table` belong to one of these categories."""

    COLUMN_HEADER = 0
    SECTION_TITLE = 1
    FLOATING_FOOTER = 2
    FLOATING_TITLE = 3
    SUMMARY_CELL = 4


FORCE_SYNC = "FORCE_SYNC"
FORCE_ASYNC = "FORCE_ASYNC"
DEFAULT = "DEFAULT"


FORMS = "FORMS"
TABLES = "TABLES"
QUERIES = "QUERIES"
LAYOUT = "LAYOUT"


class TextractFeatures(Enum):
    """Features to be used as parameter for AnalyzeDocument and StartDocumentAnalysis."""

    FORMS = 0
    TABLES = 1
    QUERIES = 2
    SIGNATURES = 3
    LAYOUT = 4


class TextractType(Enum):
    """Document Entity types recognized by Textract APIs."""

    WORDS = 0
    LINES = 1
    KEY_VALUE_SET = 2
    SELECTION_ELEMENT = 3
    TABLES = 4
    TABLE_CELL = 5


class DirectionalFinderType(Enum):
    """Document Entity types recognized by Textract APIs."""

    KEY_VALUE_SET = 0
    SELECTION_ELEMENT = 1


class TableFormat(Enum):
    """Various formats of printing :class:`Table` data with the tabulate package."""

    CSV = 0
    PLAIN = 1
    SIMPLE = 2
    GITHUB = 3
    GRID = 4
    FANCY_GRID = 5
    PIPE = 6
    ORGTBL = 7
    JIRA = 8
    PRESTO = 9
    PRETTY = 10
    PSQL = 11
    RST = 12
    MEDIAWIKI = 13
    MOINMOIN = 14
    YOUTRACK = 15
    HTML = 16
    UNSAFEHTML = 17
    LATEX = 18
    LATEX_RAW = 19
    LATEX_BOOKTABS = 20
    LATEX_LONGTABLE = 21
    TEXTILE = 22
    TSV = 23


class SimilarityMetric(Enum):
    """
    Similarity metrics for search queries on Document data

    COSINE: Cosine similarity is a metric used to measure the similarity of two vectors. It measures the similarity in the direction or orientation of the vectors ignoring differences in their magnitude or scale.
    EUCLIDEAN: Euclidean distance is calculated as the square root of the sum of the squared differences between the two vectors.
    LEVENSHTEIN: The Levenshtein distance is a string metric for measuring difference between two sequences. It is the minimum number of single-character edits (i.e. insertions, deletions or substitutions) required to change one word into the other.
    """

    COSINE = 0
    EUCLIDEAN = 1
    LEVENSHTEIN = 2


class TextractAPI(Enum):
    """API types for asynchronous type fetching"""

    DETECT_TEXT = 0
    ANALYZE = 1
    EXPENSE = 2

    @classmethod
    def TextractAPI_to_Textract_API(cls, api):
        if api == TextractAPI.DETECT_TEXT:
            return Textract_API.DETECT
        elif api == TextractAPI.ANALYZE:
            return Textract_API.ANALYZE
        elif api == TextractAPI.EXPENSE:
            return Textract_API.EXPENSE
        else:
            raise InputError()

    @classmethod
    def Textract_API_to_TextractAPI(cls, api: Textract_API):
        if api == Textract_API.DETECT:
            return TextractAPI.DETECT_TEXT
        elif api == Textract_API.ANALYZE:
            return TextractAPI.ANALYZE
        elif api == Textract_API.EXPENSE:
            return TextractAPI.EXPENSE
        else:
            raise InputError()


class AnalyzeIDFields(Enum):
    """Enum containing all the AnalyzeID keys"""

    FIRST_NAME = "FIRST_NAME"
    LAST_NAME = "LAST_NAME"
    MIDDLE_NAME = "MIDDLE_NAME"
    SUFFIX = "SUFFIX"
    CITY_IN_ADDRESS = "CITY_IN_ADDRESS"
    ZIP_CODE_IN_ADDRESS = "ZIP_CODE_IN_ADDRESS"
    STATE_IN_ADDRESS = "STATE_IN_ADDRESS"
    COUNTY = "COUNTY"
    DOCUMENT_NUMBER = "DOCUMENT_NUMBER"
    EXPIRATION_DATE = "EXPIRATION_DATE"
    DATE_OF_BIRTH = "DATE_OF_BIRTH"
    STATE_NAME = "STATE_NAME"
    DATE_OF_ISSUE = "DATE_OF_ISSUE"
    CLASS = "CLASS"
    RESTRICTIONS = "RESTRICTIONS"
    ENDORSEMENTS = "ENDORSEMENTS"
    ID_TYPE = "ID_TYPE"
    VETERAN = "VETERAN"
    ADDRESS = "ADDRESS"
    # Only available in passports
    PLACE_OF_BIRTH = "PLACE_OF_BIRTH"


class AnalyzeExpenseLineItemFields(Enum):
    ITEM = "ITEM"
    PRICE = "PRICE"
    PRODUCT_CODE = "PRODUCT_CODE"
    QUANTITY = "QUANTITY"
    UNIT_PRICE = "UNIT_PRICE"
    EXPENSE_ROW = "EXPENSE_ROW"


class AnalyzeExpenseFields(Enum):
    ACCOUNT_NUMBER = "ACCOUNT_NUMBER"
    ADDRESS = "ADDRESS"
    ADDRESS_BLOCK = "ADDRESS_BLOCK"
    AMOUNT_DUE = "AMOUNT_DUE"
    AMOUNT_PAID = "AMOUNT_PAID"
    CITY = "CITY"
    COUNTRY = "COUNTRY"
    CUSTOMER_NUMBER = "CUSTOMER_NUMBER"
    DELIVERY_DATE = "DELIVERY_DATE"
    DISCOUNT = "DISCOUNT"
    DUE_DATE = "DUE_DATE"
    GRATUITY = "GRATUITY"
    INVOICE_RECEIPT_DATE = "INVOICE_RECEIPT_DATE"
    INVOICE_RECEIPT_ID = "INVOICE_RECEIPT_ID"
    NAME = "NAME"
    ORDER_DATE = "ORDER_DATE"
    OTHER = "OTHER"
    PAYMENT_TERMS = "PAYMENT_TERMS"
    PO_NUMBER = "PO_NUMBER"
    PRIOR_BALANCE = "PRIOR_BALANCE"
    RECEIVER_ABN_NUMBER = "RECEIVER_ABN_NUMBER"
    RECEIVER_ADDRESS = "RECEIVER_ADDRESS"
    RECEIVER_GST_NUMBER = "RECEIVER_GST_NUMBER"
    RECEIVER_NAME = "RECEIVER_NAME"
    RECEIVER_PAN_NUMBER = "RECEIVER_PAN_NUMBER"
    RECEIVER_PHONE = "RECEIVER_PHONE"
    RECEIVER_VAT_NUMBER = "RECEIVER_VAT_NUMBER"
    SERVICE_CHARGE = "SERVICE_CHARGE"
    SHIPPING_HANDLING_CHARGE = "SHIPPING_HANDLING_CHARGE"
    STATE = "STATE"
    STREET = "STREET"
    SUBTOTAL = "SUBTOTAL"
    TAX = "TAX"
    TAX_PAYER_ID = "TAX_PAYER_ID"
    TOTAL = "TOTAL"
    VENDOR_ABN_NUMBER = "VENDOR_ABN_NUMBER"
    VENDOR_ADDRESS = "VENDOR_ADDRESS"
    VENDOR_GST_NUMBER = "VENDOR_GST_NUMBER"
    VENDOR_NAME = "VENDOR_NAME"
    VENDOR_PAN_NUMBER = "VENDOR_PAN_NUMBER"
    VENDOR_PHONE = "VENDOR_PHONE"
    VENDOR_URL = "VENDOR_URL"
    VENDOR_VAT_NUMBER = "VENDOR_VAT_NUMBER"
    ZIP_CODE = "ZIP_CODE"


class AnalyzeExpenseFieldsGroup(Enum):
    RECEIVER = "RECEIVER"
    RECEIVER_BILL_TO = "RECEIVER_BILL_TO"
    RECEIVER_SHIP_TO = "RECEIVER_SHIP_TO"
    RECEIVER_SOLD_TO = "RECEIVER_SOLD_TO"
    VENDOR = "VENDOR"
    VENDOR_REMIT_TO = "VENDOR_REMIT_TO"
    VENDOR_SUPPLIER = "VENDOR_SUPPLIER"


class CLIPrint(Enum):
    ALL = 0
    TEXT = 1
    TABLES = 2
    FORMS = 3
    QUERIES = 4
    EXPENSES = 5
    SIGNATURES = 6
    IDS = 7
    LAYOUTS = 8

class CLIOverlay(Enum):
    ALL = 0
    WORDS = 1
    LINES = 2
    TABLES = 3
    FORMS = 4
    QUERIES = 5
    SIGNATURES = 6
    LAYOUTS = 7



import statistics
from typing import List
from copy import deepcopy
from collections.abc import Iterable


def flatten(list_of_lists):
    """
    Utility function to flatten a list of lists recursively.

    :param list_of_lists: List containing any depth of lists recursively to be flattened into a single list.
    :type list_of_lists: list
    :return: Flattened list of input list
    :rtype: list
    """
    for x in list_of_lists:
        if isinstance(x, Iterable):
            yield from flatten(x)
        else:
            yield x


def get_indices(numpy_indexing: str = ":", max_val=10) -> List[int]:
    """
    Function to convert numpy indexing format to list of indices to access cells within the Table.

    :param numpy_indexing: string containing start:stop:step format
    :param max_val: maximum rows or columns on the table depending on input.
    :return: Returns the indices of table rows and columns following the numpy indexing format.
    :rtype: list
    """
    indices = []
    assert isinstance(numpy_indexing, str)
    assert ":" in numpy_indexing or numpy_indexing.isdigit()

    if numpy_indexing == ":":
        numpy_indexing = "None:None"
    if numpy_indexing == "None":
        numpy_indexing = "None:None"

    if ":" not in numpy_indexing:
        if int(numpy_indexing) > max_val:
            raise IndexError()
        return [int(numpy_indexing)]

    if numpy_indexing == "None:None:None":
        indices = list(range(0, max_val))

    else:
        return_indices = numpy_indexing.split(":")
        assert len(return_indices) > 1

        start = (
            int(return_indices[0])
            if return_indices[0] != "" and return_indices[0] != "None"
            else 0
        )

        if start < 0:
            start = max_val + start

        end = (
            int(return_indices[1])
            if return_indices[1] != "" and return_indices[1] != "None"
            else max_val
        )

        if end < 0:
            end = max_val + end

        index_range = list(range(start, end))

        if len(return_indices) == 3:
            step = (
                int(return_indices[2])
                if return_indices[2] != "" and return_indices[2] != "None"
                else 1
            )
            indices += [i for i in index_range if index_range.index(i) % step == 0]
        else:
            indices = index_range

    return list(set(indices))


def sort_by_position(entities: List) -> List:
    return sorted(entities, key=lambda e: (e.bbox.y + e.bbox.height, e.bbox.x))



import re
from typing import List, Union
from PIL import Image

try:
    import pypdfium2
    PYPDFIUM2_IS_INSTALLED = True
except ImportError:
    PYPDFIUM2_IS_INSTALLED = False

try:
    from pdf2image import convert_from_bytes, convert_from_path, pdfinfo_from_bytes, pdfinfo_from_path
    PDF2IMAGE_IS_INSTALLED = True
except ImportError:
    PDF2IMAGE_IS_INSTALLED = False
    

def rasterize_pdf(pdf: Union[str, bytes]) -> List[Image.Image]:
    """
    Convert a pdf into a list of images
    """
    if PYPDFIUM2_IS_INSTALLED:
        pdf = pypdfium2.PdfDocument(pdf)
        return [page.render(scale=250 / 72).to_pil() for page in pdf]
    elif PDF2IMAGE_IS_INSTALLED:
        if isinstance(pdf, str):
            return convert_from_path(pdf, dpi=250, fmt="jpeg")
        elif isinstance(pdf, bytes):
            return convert_from_bytes(pdf, dpi=250, fmt="jpeg")
        else:
            raise Exception(f"{type(pdf)} is not a supported type, should be str or bytes")
    else:
        raise Exception("PDF rasterization is not possible if neither pypdfium2 nor pdf2image are installed")


class IdentityField:
    def __init__(self, key, value, confidence):
        self._key = key
        self._value = value
        self._confidence = confidence

    @property
    def key(self) -> str:
        return self._key

    @property
    def value(self) -> str:
        return self._value

    @property
    def confidence(self) -> float:
        return self._confidence

    def __repr__(self) -> str:
        return self.value



"""
Represents a single :class:`Line` Entity within the :class:`Document`. 
The Textract API response returns groups of words as LINE BlockTypes. They contain :class:`Word` entities as children. 

This class contains the associated metadata with the :class:`Line` entity including the entity ID, 
bounding box information, child words, page number, Page ID and confidence of detection.
"""

import logging
from typing import List

from textractor.entities.word import Word
from textractor.data.constants import TextTypes
from textractor.entities.bbox import BoundingBox
from textractor.exceptions import InputError
from textractor.entities.document_entity import DocumentEntity
from textractor.visualizers.entitylist import EntityList


class Line(DocumentEntity):
    """
    To create a new :class:`Line` object we need the following:

    :param entity_id: Unique identifier of the Line entity.
    :type entity_id: str
    :param bbox: Bounding box of the line entity.
    :type bbox: BoundingBox
    :param words: List of the Word entities present in the line
    :type words: list, optional
    :param confidence: confidence with which the entity was detected.
    :type confidence: float, optional
    """

    def __init__(
        self,
        entity_id: str,
        bbox: BoundingBox,
        words: List[Word] = None,
        confidence: float = 0,
    ):
        super().__init__(entity_id, bbox)
        if words is not None and len(words) > 0:
            self._children: List[Word] = sorted(words, key=lambda x: (x.bbox.x, x.bbox.y))
        else:
            self._children = []

        self._confidence = confidence / 100
        self._page = None
        self._page_id = None

    @property
    def text(self):
        """
        :return: Returns the text transcription of the :class:`Line` entity.
        :rtype: str
        """
        return " ".join([word.text for word in self.words])

    @property
    def words(self):
        """
        :return: Returns the line's children
        :rtype: List[Word]
        """
        return self._children

    def get_text_and_words(self, config):
        if not self.bbox:
            self.bbox = BoundingBox.enclosing_bbox(self.words)
        for w in self.words:
            w.line_id = self.id
            w.line_bbox = self.bbox
        return self.text, self.words

    @property
    def page(self):
        """
        :return: Returns the page number of the page the :class:`Line` entity is present in.
        :rtype: int
        """
        return self._page

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the Line entity.

        :param page_num: Page number where the Line entity exists.
        :type page_num: int
        """
        self._page = page_num

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """
        return self._page_id

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the :class:`Line` entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """
        self._page_id = page_id

    def get_words_by_type(self, text_type: TextTypes = TextTypes.PRINTED) -> List[Word]:
        """
        :param text_type: TextTypes.PRINTED or TextTypes.HANDWRITING
        :type text_type: TextTypes
        :return: Returns EntityList of Word entities that match the input text type.
        :rtype: EntityList[Word]
        """
        if not isinstance(text_type, TextTypes):
            raise InputError(
                "text_type parameter should be of TextTypes type. Find input choices from textractor.data.constants"
            )

        if not self.words:
            return []
        return EntityList([word for word in self.words if word.text_type == text_type])

    def __repr__(self):
        """
        :return: String representation of the Line entity.
        :rtype: str
        """
        return " ".join([word.text for word in self.words])
